#
# Copyright 2015-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import os

import pytest

from numina.tests.testcache import download_cache
from numina.core import ObservationResult
from numina.core import DataFrame
from numina.instrument.assembly import assembly_instrument

from megaradrp.recipes.calibration.bias import BiasRecipe
from megaradrp.loader import load_drp


@pytest.mark.remote_data
def test_bias():

    BASE_URL = 'http://guaix.fis.ucm.es/~spr/megara_test/BIAS/%s'
    images = ['e99d2937d2c29a27c0ba4eebfcf7918e',
              'e99d2937d2c29a27c0ba4eebfcf7918e',
              'e99d2937d2c29a27c0ba4eebfcf7918e']

    fs = [download_cache(BASE_URL % i) for i in images]

    ob = ObservationResult()
    ob.instrument = 'MEGARA'
    ob.mode = 'MegaraBiasImage'
    ob.frames = [DataFrame(filename=f.name) for f in fs]

    insdrp = load_drp()
    pipeline = insdrp.pipelines.get('default')
    recipe = pipeline.get_recipe_object(ob.mode)

    assert isinstance(recipe, BiasRecipe)
    # TODO: these should be created by the DAL

    import numina.instrument.assembly as asbl
    import numina.drps
    sys_drps = numina.drps.get_system_drps()
    com_store = asbl.load_panoply_store(sys_drps)

    key, date_obs, keyname = insdrp.select_profile(ob)
    ob.configuration = assembly_instrument(com_store, key, date_obs, by_key=keyname)

    ri = recipe.create_input(obresult=ob)

    result = recipe.run(ri)
    # assert result.qc >= QC.UNKNOWN

    # Checks on the image
    hdulist = result.master_bias.open()
    assert len(hdulist) == 1

    hdu = hdulist[0]
    assert hdu.shape == (4112, 4096)

    data = hdu.data
    mlevel = 0.0

    block = data[1980:2020, 1980:2020]
    mblock = block.mean()
    sblock = block.std()

    assert abs(mblock - mlevel) < 5 * sblock

    # In the end, remove the files
    for f in fs:
        os.remove(f.name)



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Tests for the calibration module."""

import pytest

from numina.user.cli import main
import numina.drps
from megaradrp.recipes.calibration.bias import BiasRecipe
from megaradrp.loader import load_drp

BASE_URL = 'http://guaix.fis.ucm.es/~spr/megara_test/'


def run_recipe():
    main(['run', 'obsrun.yaml', '-r', 'control.yaml'])


def test_recipe1(drpmocker):

    drpmocker.add_drp('MEGARA', load_drp)

    insdrp = numina.drps.get_system_drps().query_by_name('MEGARA')
    pipeline = insdrp.pipelines.get('default')
    recipe = pipeline.get_recipe_object('MegaraBiasImage')

    assert isinstance(recipe, BiasRecipe)


@pytest.mark.remote_data
@pytest.mark.usefixtures("numinatpldir")
def test_mode_bias_set0(drpmocker):

    drpmocker.add_drp('MEGARA', load_drp)

    run_recipe()


@pytest.mark.remote_data
@pytest.mark.usefixtures("numinatpldir")
def test_mode_bias_set1(drpmocker):

    drpmocker.add_drp('MEGARA', load_drp)

    run_recipe()


@pytest.mark.remote_data
@pytest.mark.usefixtures("numinatpldir")
def test_mode_trace_map_set0(drpmocker):

    drpmocker.add_drp('MEGARA', load_drp)
    run_recipe()


@pytest.mark.remote_data
@pytest.mark.usefixtures("numinatpldir")
def test_mode_fiber_flat_set0(drpmocker):

    drpmocker.add_drp('MEGARA', load_drp)
    run_recipe()



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import datetime

import numpy
import numpy.polynomial.polynomial as nppol
import astropy.io.fits as fits
import numina.array.trace.extract as extract
import numina.processing


_logger = logging.getLogger(__name__)


def apextract(data, trace):
    """Extract apertures."""
    rss = numpy.empty((trace.shape[0], data.shape[1]), dtype='float32')
    for idx, r in enumerate(trace):
        l = r[0]
        r = r[2] + 1
        sl = (slice(l, r), )
        m = data[sl].sum(axis=0)
        rss[idx] = m
    return rss


def apextract_tracemap(data, tracemap):
    """Extract apertures using a tracemap.

    Consider that the nearest fiber could be far away if there
    are missing fibers.

    Parameters
    ----------

    data: ndarray
    tracemap: TraceMap

    """

    existing = [None]
    for t in tracemap.contents:
        if t.valid:
            existing.append(t)

    existing.append(None)
    # Compute borders
    borders = []
    far_dist = 100

    # Handle the first and last using centinels
    for t1, t2, t3 in zip(existing, existing[1:], existing[2:]):
        # Distance to contfibers # in box
        if t1 is None:
            d21 = far_dist
        else:
            d21 = (t2.fibid - t1.fibid) + (t2.boxid - t1.boxid)
        if t3 is None:
            d32 = far_dist
        else:
            d32 = (t3.fibid - t2.fibid) + (t3.boxid - t2.boxid)

        y2_off = t2.polynomial(tracemap.ref_column)
        t2_off = tracemap.global_offset(y2_off)
        t2_pol_off = t2.polynomial + t2_off

        # Right border
        pix_32 = None
        if d32 <= 2:
            y3_off = t3.polynomial(tracemap.ref_column)
            t3_off = tracemap.global_offset(y3_off)
            pix_32 = 0.5 * (t2_pol_off + t3.polynomial + t3_off)
            if d32 == 2:
                pix_32 = 0.5 * (pix_32 + t2_pol_off)
        elif d32 > 2:
            pix_32 = None

        pix_21 = None
        if d21 <= 2:
            y1_off = t1.polynomial(tracemap.ref_column)
            t1_off = tracemap.global_offset(y1_off)
            pix_21 = 0.5 * (t2_pol_off + t1.polynomial + t1_off)
            if d21 == 2:
                pix_21 = 0.5 * (pix_21 + t2_pol_off)
        elif d21 > 2:
            pix_21 = None

        if pix_32 is None and pix_21 is None:
            continue

        if pix_32 is None:
            # Recompute pix32 using pix_21
            pix_32 = t2_pol_off + (t2_pol_off - pix_21)

        if pix_21 is None:
            # Recompute pix21 using pix_32
            pix_21 = t2_pol_off - (pix_32 - t2_pol_off)

        borders.append((t2.fibid, pix_21, pix_32))

    nfibers = tracemap.total_fibers
    out = numpy.zeros((nfibers, data.shape[1]), dtype='float')
    rss = extract_simple_rss2(data, borders, out=out)

    return rss


def extract_simple_rss(arr, borders2, axis=0, out=None):

    # FIXME, this should be changed in numina
    # If arr is not in native byte order, the C-extension won't work
    if arr.dtype.byteorder != '=':
        arr2 = arr.byteswap().newbyteorder()
    else:
        arr2 = arr

    if axis == 0:
        arr3 = arr2
    elif axis == 1:
        arr3 = arr2.T
    else:
        raise ValueError("'axis' must be 0 or 1")

    if out is None:
        out = numpy.zeros((borders2[-1][0], arr3.shape[1]), dtype='float')

    xx = numpy.arange(arr3.shape[1])

    # Borders contains a list of function objects
    for idx, b1, b2 in borders2:
        bb1 = b1(xx)
        bb1[bb1 < -0.5] = -0.5
        bb2 = b2(xx)
        bb2[bb2 > arr3.shape[0] - 0.5] = arr3.shape[0] - 0.5
        extract.extract_simple_intl(arr3, xx, bb1, bb2, out[idx-1])
    return out


apextract_tracemap_2 = apextract_tracemap
extract_simple_rss2 = extract_simple_rss


class ApertureExtractor(numina.processing.Corrector):
    """A Node that extracts apertures."""

    def __init__(self, trace_repr, datamodel=None, dtype='float32',
                 processes=0, offset=None):

        if offset:
            trace_repr.global_offset = trace_repr.global_offset + nppol.Polynomial(offset)

        self.trace_repr = trace_repr
        self.processes = processes
        super(ApertureExtractor, self).__init__(
            datamodel=datamodel,
            calibid=trace_repr.uuid,
            dtype=dtype
        )

    def run(self, img):
        # workaround
        imgid = self.get_imgid(img)

        method_name = 'simple'
        simple = True
        if hasattr(self.trace_repr, 'aper_extract'):
            simple = False
            method_name = 'advanced'

        if simple:
            _logger.debug('simple aperture extraction')
            _logger.debug('extracting (apextract_tracemap) in image %s', imgid)
            _logger.debug('with trace map %s', self.calibid)
        else:
            _logger.debug('advanced aperture extraction')
            _logger.debug('extracting (apextract_model) in image %s', imgid)
            _logger.debug('with model map %s', self.calibid)

        _logger.debug('offsets are %s', self.trace_repr.global_offset.coef)
        if simple:
            rssdata = apextract_tracemap(img[0].data, self.trace_repr)
        else:
            rssdata = self.trace_repr.aper_extract(img[0].data, processes=self.processes)

        img[0].data = rssdata

        hdr = img[0].header

        hdr['NUM-APE'] = self.calibid
        hdr['history'] = 'Aperture extraction method {}'.format(method_name)
        hdr['history'] = 'Aperture extraction with {}'.format(self.calibid)
        hdr['history'] = 'Aperture extraction offsets are {}'.format(self.trace_repr.global_offset.coef.tolist())
        hdr['history'] = 'Aperture extraction time {}'.format(datetime.datetime.utcnow().isoformat())

        # Update Fibers
        fibers_ext = img['FIBERS']
        fibers_ext_headers = fibers_ext.header
        for aper in self.trace_repr.contents:
            key = "FIB%03d_V" % aper.fibid
            fibers_ext_headers[key] = aper.valid
            key = "FIB%03dS1" % aper.fibid
            fibers_ext_headers[key] = aper.start
            key = "FIB%03dS2" % aper.fibid
            fibers_ext_headers[key] = aper.stop

        newimg = fits.HDUList([img[0], fibers_ext])
        return newimg



#
# Copyright 2018-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import math

import numpy

from megaradrp.instrument import MEGARA_IAA


def compute_pa_from_ipa(ipa, iaa=MEGARA_IAA):
    """Recompute the PA from IPA

    Parameters
    ==========
    ipa: float
        Instrument Position Angle
    iaa: float
        Instrument Alignment Angle

    Returns
    =======
    position angle

    Compute the Position angle of the image from the IPA and
    the (generally fixed) Instrument Alignment Angle.
    The angles must be in the same units.

    """

    pa = -iaa + ipa
    return pa


def update_wcs_from_ipa(hdr, pa):
    """Recompute the WCS rotations from PA"""
    pa_rad = numpy.deg2rad(pa)
    cos_pa = math.cos(pa_rad)
    sin_pa = math.sin(pa_rad)

    # Update PC_ keywords
    hdr['PC1_1'] = cos_pa
    hdr['PC2_2'] = cos_pa
    hdr['PC1_2'] = sin_pa
    hdr['PC2_1'] = -sin_pa
    # CDELT1 must be negative
    hdr['CDELT1'] = -abs(hdr['CDELT1'])

    return hdr


#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""MOS Standard Star Image Recipe for Megara"""


from scipy.interpolate import interp1d
import astropy.io.fits as fits
import astropy.units as u
from astropy import constants as const

from numina.core import Result, Parameter
from numina.core.requirements import Requirement
from numina.types.array import ArrayType

from megaradrp.processing.extractobj import extract_star, generate_sensitivity
from megaradrp.recipes.scientific.base import ImageRecipe
from megaradrp.types import ProcessedRSS, ProcessedFrame, ProcessedSpectrum
from megaradrp.types import ReferenceSpectrumTable, ReferenceExtinctionTable
from megaradrp.types import MasterSensitivity


class MOSStandardRecipe(ImageRecipe):
    """Process MOS Standard Star Recipe.

    This recipe processes a set of images
    obtained in **MOS Stardard Star image** mode and returns
    the total flux of the star.

    See Also
    --------
    megaradrp.recipes.calibration.lcbstdstar.LCBStandardRecipe

    Notes
    -----
    Images provided by `obresult` are trimmed and corrected
    from overscan, bad pixel mask (if `master_bpm` is not None),
    bias, dark current (if `master_dark` is not None) and
    slit-flat (if `master_slitflat` is not None).

    Images thus corrected are then stacked using the median.
    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    The sky is subtracted by combining the the fibers marked as `SKY`
    in the fibers configuration. The RSS with sky subtracted is returned ini the
    field `final_rss` of the recipe result.

    The flux of the star is computed by adding the 7 fibers corresponding to the bundle
    containing the star and returned as `star_spectrum`.

    """
    position = Requirement(list, "Position of the reference object", default=(0, 0))
    # nrings = 1
    reference_spectrum = Requirement(ReferenceSpectrumTable, "Spectrum of reference star")
    reference_spectrum_velocity = Parameter(0.0, 'Radial velocity of reference spectrum')
    reference_extinction = Requirement(ReferenceExtinctionTable, "Reference extinction")
    sigma_resolution = Parameter(20.0, 'sigma Gaussian filter to degrade resolution ')

    reduced_image = Result(ProcessedFrame)
    final_rss = Result(ProcessedRSS)
    reduced_rss = Result(ProcessedRSS)
    sky_rss = Result(ProcessedRSS)
    star_spectrum = Result(ProcessedSpectrum)
    master_sensitivity = Result(MasterSensitivity)
    fiber_ids = Result(ArrayType)

    def run(self, rinput):

        self.logger.info('starting MOSStandardRecipe reduction')

        reduced2d, rss_data = super(MOSStandardRecipe, self).base_run(rinput)

        self.logger.info('start sky subtraction')
        final, origin, sky = self.run_sky_subtraction(rss_data)
        self.logger.info('end sky subtraction')

        # 1 + 6  for first ring
        # 1 + 6  + 12  for second ring
        # 1 + 6  + 12  + 18 for third ring
        # 1 + 6 * Sum_i=0^n =  1 + 3 * n * (n +1)
        # In MOS, only 1 ring around central point
        self.logger.debug('adding %d nrings', 1)
        npoints = 7
        self.logger.debug('adding %d fibers', npoints)

        fiberconf = self.datamodel.get_fiberconf(final)
        spectra_pack = extract_star(final, rinput.position, npoints,
                                    fiberconf, logger=self.logger)

        spectrum, colids, cover1, cover2 = spectra_pack
        star_spectrum = fits.PrimaryHDU(spectrum, header=final[0].header)

        rad_vel = rinput.reference_spectrum_velocity * u.km / u.s
        factor = 1 + rad_vel / const.c
        star_interp = interp1d(rinput.reference_spectrum[:, 0] / factor,
                               rinput.reference_spectrum[:, 1])
        
        extinc_interp = interp1d(rinput.reference_extinction[:, 0],
                                 rinput.reference_extinction[:, 1])

        fiber_ids = [colid + 1 for colid in colids]
        sigma = rinput.sigma_resolution
        sens = generate_sensitivity(final, spectrum, star_interp, extinc_interp, cover1, cover2, sigma)
        self.logger.info('end MOSStandardRecipe reduction')

        return self.create_result(
            reduced_image=reduced2d,
            final_rss=final,
            reduced_rss=origin,
            sky_rss=sky,
            star_spectrum=star_spectrum,
            master_sensitivity=sens,
            fiber_ids=fiber_ids
        )



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""MOS Image Recipe for Megara"""

import numpy
from scipy.interpolate import interp1d
import astropy.wcs
import astropy.units as u

from numina.core import Result

from megaradrp.processing.fluxcalib import FluxCalibration
from megaradrp.utils import add_collapsed_mos_extension
from megaradrp.types import ProcessedRSS, ProcessedFrame
from .base import ImageRecipe


class MOSImageRecipe(ImageRecipe):
    """Process MOS images.

    This recipe processes a set of images
    obtained in **MOS image** mode and returns
    the sky subtracted RSS.

    See Also
    --------
    megaradrp.recipes.scientific.lcb.LCBImageRecipe

    Notes
    -----
    Images provided by `obresult` are trimmed and corrected
    from overscan, bad pixel mask (if `master_bpm` is not None),
    bias, dark current (if `master_dark` is not None) and
    slit-flat (if `master_slitflat` is not None).

    Images thus corrected are the stacked using the median.
    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    The sky is subtracted by combining the the fibers marked as `SKY`
    in the fibers configuration. The RSS with sky subtracted is returned in the
    field `final_rss` of the recipe result.

    If a `master_sensitivity` is provided (optional), RSS products will be
    flux calibrated. If `reference_extinction` is provided (optional),
    `final_rss` and `reduced_rss` will be extinction corrected. Notice that
    `sky_rss` is not corrected from extinction.

    """

    reduced_image = Result(ProcessedFrame)
    final_rss = Result(ProcessedRSS)
    reduced_rss = Result(ProcessedRSS)
    sky_rss = Result(ProcessedRSS)

    def run(self, rinput):
        self.logger.info('starting MOS reduction')

        reduced2d, rss_data = super(MOSImageRecipe, self).base_run(rinput)

        self.logger.info('start sky subtraction')
        isb = rinput.ignored_sky_bundles
        if isb:
            self.logger.info('sky bundles ignored: %s', isb)
        final, origin, sky = self.run_sky_subtraction(rss_data, ignored_sky_bundles=isb)
        self.logger.info('end sky subtraction')
        # Flux calibration
        if rinput.master_sensitivity is not None:
            self.logger.info('start flux calibration')
            node = FluxCalibration(rinput.master_sensitivity.open(), self.datamodel)
            final = node(final)
            origin = node(origin)
            sky = node(sky)
            self.logger.info('end flux calibration')
        else:
            self.logger.info('no flux calibration')

        # Extinction calibration
        if rinput.reference_extinction is not None:
            self.logger.info('start extinction correction')
            extinc_interp = interp1d(rinput.reference_extinction[:, 0],
                                     rinput.reference_extinction[:, 1])

            wlcalib = astropy.wcs.WCS(final[0].header)
            pixrange = numpy.arange(final[0].data.shape[1])
            yrange = pixrange * 0
            calcwl = numpy.array([pixrange, yrange]).T
            wavelen_ = wlcalib.all_pix2world(calcwl, 0.0)
            if wlcalib.wcs.cunit[0] == u.dimensionless_unscaled:
                # CUNIT is empty, assume Angstroms
                wavelen = wavelen_[:, 0] * u.AA
            else:
                wavelen = wavelen_[:, 0] * wlcalib.wcs.cunit[0]
            wavelen_aa = wavelen.to(u.AA).value

            airmass = final[0].header['AIRMASS']

            extinc_corr = numpy.power(10.0, 0.4 * extinc_interp(wavelen_aa) * airmass)

            final[0].data *= extinc_corr
            origin[0].data *= extinc_corr
            # sky is not corrected from extinction
            self.logger.info('end extinction correction')
        else:
            self.logger.info('no extinction correction')

        self.logger.info('add collapsed extension')
        final = add_collapsed_mos_extension(final)
        origin = add_collapsed_mos_extension(origin)

        self.logger.info('end MOS reduction')
        return self.create_result(
            reduced_image=reduced2d,
            final_rss=final,
            reduced_rss=origin,
            sky_rss=sky
        )








from numina.core import BaseRecipe

from megaradrp.core.recipe import MegaraBaseRecipe


def test_base_recipe():
    version = "1.0.1"
    obj = MegaraBaseRecipe(version=version)
    assert isinstance(obj, BaseRecipe)
    assert obj.__version__ == version


#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


class FiberBundle(object):
    def __init__(self, name, bid, static=True):
        super(FiberBundle, self).__init__()
        # Geometry of the fibers
        self.name = name
        self.bunds_id = bid
        self.static = static
        self.lf = []

    def add_light_fiber(self, lf):
        self.lf.append(lf)

    @property
    def fibs_id(self):
        return [ch.fibid for ch in self.lf]

    @property
    def inactive_fibs_id(self):
        return [ch.fibid for ch in self.lf if ch.inactive]

    @property
    def active_fibs_id(self):
        return [ch.fibid for ch in self.lf if not ch.inactive]

    def config_info(self):
        return {'name': self.name,
                'nfibers': len(self.lf),
                'fibs_id': self.fibs_id,
                'id': self.bunds_id,
                'static': self.static,
                'inactive_fibs_id': self.inactive_fibs_id
                }


#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy


class FocalPlane(object):

    def __init__(self, cover):

        self.focalbundle = {}
        self.cover = cover
        self.ddtype = [('fibid', 'i4'),('x', 'f4'), ('y', 'f4'), ('cover', 'f4')]

    def set_cover(self, mode):
        """Cover in the focal plane."""

        self.cover.set_mode(mode)

    def connect_lcb(self, lcb):
        self.focalbundle['LCB'] = lcb

    def connect_fibermos(self, mos):
        self.focalbundle['MOS'] = mos

    def get_visible_fibers(self, name):
        fibid, allpos = self.focalbundle[name].fibers_in_focal_plane()

        p1 = self.cover.visible_fibers(fibid, allpos)

        return numpy.array(p1, dtype=self.ddtype)

    def filter_visible_fibers(self, fibid, allpos):

        p1 = self.cover.visible_fibers(fibid, allpos)

        return numpy.array(p1, dtype=self.ddtype)

    def get_all_fibers(self, name):

        fibid, all_pos = self.focalbundle[name].fibers_in_focal_plane()

        p2 = numpy.empty((len(fibid,)), dtype=self.ddtype)
        p2['fibid'] = fibid
        p2['x'] = all_pos[:,0]
        p2['y'] = all_pos[:,1]
        p2['cover'] = 1.0

        return p2



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging

from numina.core import BaseRecipe
from numina.core import DataFrame
from numina.types.qc import QC
from numina.core.requirements import ObservationResultRequirement
import numina.util.flow as flowmod


import megaradrp.core.correctors as cor
from megaradrp.datamodel import MegaraDataModel


class MegaraBaseRecipe(BaseRecipe):
    """Base clase for all MEGARA Recipes


    Parameters
    ----------
    intermediate_results : bool, optional
                           If True, save intermediate results of the Recipe


    Attributes
    ----------

    obresult : ObservationResult, requirement
    logger :
         recipe logger

    datamodel : MegaraDataModel

    """

    obresult = ObservationResultRequirement()
    logger = logging.getLogger('numina.recipes.megara')
    datamodel = MegaraDataModel()

    def validate_input(self, recipe_input):
        """Method to customize recipe input validation.

        See Also
        --------
        numina.core.validator.validate

        """
        self.logger.info('start validating input')
        super(MegaraBaseRecipe, self).validate_input(recipe_input)
        self.logger.info('end validating input')

    def run_qc(self, recipe_input, recipe_result):
        """Run Quality Control checks."""
        recipe_result.qc = QC.GOOD
        return recipe_result

    def types_getter(self):
        from megaradrp.types import MasterBias, MasterDark, MasterBPM, MasterSlitFlat
        imgtypes = [MasterBPM, MasterBias, MasterDark, MasterSlitFlat]
        getters = [cor.get_corrector_bpm, cor.get_corrector_bias,
                   [cor.get_corrector_dark, cor.get_corrector_gain],
                   cor.get_corrector_slit_flat
                   ]
        return imgtypes, getters

    def get_filters(self, imgtypes, getters):
        import collections
        used_getters = []
        for rtype, getter in zip(imgtypes, getters):
            self.logger.debug('get_filters, %s  %s', rtype, getter)
            if rtype is None:
                # Unconditional
                if isinstance(getter, collections.Iterable):
                    used_getters.extend(getter)
                else:
                    used_getters.append(getter)
            else:
                # Search
                for key, val in self.RecipeInput.stored().items():
                    if isinstance(val.type, rtype):
                        if isinstance(getter, collections.Iterable):
                            used_getters.extend(getter)
                        else:
                            used_getters.append(getter)
                        break
                else:
                    pass
        return used_getters

    def init_filters_generic(self, rinput, getters_seq, ins):

        meta = self.gather_info(rinput)
        self.logger.debug('obresult info')
        for entry in meta['obresult']:
            self.logger.debug('frame info is %s', entry)

        reduction_flows = []
        for getters in getters_seq:
            correctors = [getter(rinput, meta, ins, self.datamodel) for getter in getters]
            reduction_flow = flowmod.SerialFlow(correctors)
            reduction_flows.append(reduction_flow)

        return reduction_flows

    def init_filters(self, rinput, ins):
        imgtypes1 = [None]
        getters1 = [[cor.get_corrector_overscan, cor.get_corrector_trimming]]
        getters_f1 = self.get_filters(imgtypes1, getters1)
        imgtypes2, getters2 = self.types_getter()
        getters_f2 = self.get_filters(imgtypes2, getters2)

        getters_seq = [getters_f1, getters_f2]
        return self.init_filters_generic(rinput, getters_seq, ins)

    def gather_info(self, recipeinput):
        klass = recipeinput.__class__
        metadata = {}
        for key in klass.stored():
            val = getattr(recipeinput, key)
            if isinstance(val, DataFrame):
                metadata[key] = self.datamodel.gather_info_dframe(val)
            # elif isinstance(val, ObservationResult):
            elif hasattr(val, 'frames'):
                metadata[key] = self.datamodel.gather_info_oresult(val)
            else:
                pass
        return metadata



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Calibration Recipes for Megara"""

import logging

from numina.core import Product, Requirement
from numina.core.products import ArrayType
from numina.core.requirements import ObservationResultRequirement

from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.requirements import MasterBiasRequirement
from megaradrp.types import MasterFiberFlat


_logger = logging.getLogger('numina.recipes.megara')


class LCB_IFU_StdStarRecipe(MegaraBaseRecipe):

    master_bias = MasterBiasRequirement()
    obresult = ObservationResultRequirement()

    fiberflat_frame = Product(MasterFiberFlat)
    fiberflat_rss = Product(MasterFiberFlat)
    traces = Product(ArrayType)

    def run(self, rinput):
        pass


class FiberMOS_StdStarRecipe(MegaraBaseRecipe):

    master_bias = MasterBiasRequirement()
    obresult = ObservationResultRequirement()

    fiberflat_frame = Product(MasterFiberFlat)
    fiberflat_rss = Product(MasterFiberFlat)
    traces = Product(ArrayType)

    def run(self, rinput):
        pass


class SensitivityFromStdStarRecipe(MegaraBaseRecipe):

    master_bias = MasterBiasRequirement()
    obresult = ObservationResultRequirement()

    fiberflat_frame = Product(MasterFiberFlat)
    fiberflat_rss = Product(MasterFiberFlat)
    traces = Product(ArrayType)

    def run(self, rinput):
        pass


class S_And_E_FromStdStarsRecipe(MegaraBaseRecipe):

    master_bias = MasterBiasRequirement()
    obresult = ObservationResultRequirement()

    fiberflat_frame = Product(MasterFiberFlat)
    fiberflat_rss = Product(MasterFiberFlat)
    traces = Product(ArrayType)

    def run(self, rinput):
        pass



print("import: 'megaradrp'")
import megaradrp

print("import: 'megaradrp.core'")
import megaradrp.core

print("import: 'megaradrp.core.tests'")
import megaradrp.core.tests

print("import: 'megaradrp.instrument'")
import megaradrp.instrument

print("import: 'megaradrp.instrument.configs'")
import megaradrp.instrument.configs

print("import: 'megaradrp.processing'")
import megaradrp.processing

print("import: 'megaradrp.processing.tests'")
import megaradrp.processing.tests

print("import: 'megaradrp.products'")
import megaradrp.products

print("import: 'megaradrp.recipes'")
import megaradrp.recipes

print("import: 'megaradrp.recipes.auxiliary'")
import megaradrp.recipes.auxiliary

print("import: 'megaradrp.recipes.calibration'")
import megaradrp.recipes.calibration

print("import: 'megaradrp.recipes.calibration.tests'")
import megaradrp.recipes.calibration.tests

print("import: 'megaradrp.recipes.combined'")
import megaradrp.recipes.combined

print("import: 'megaradrp.recipes.scientific'")
import megaradrp.recipes.scientific

print("import: 'megaradrp.recipes.tests'")
import megaradrp.recipes.tests

print("import: 'megaradrp.simulation'")
import megaradrp.simulation

print("import: 'megaradrp.tests'")
import megaradrp.tests




#
# Copyright 2016-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Extended multiwavelength simulation"""


from __future__ import print_function


import numpy as np


def create_dummy_spectrum(wl_in):
    wlc = [0.37, 0.38, 0.39, 0.40, 0.41, 0.42]
    s = 8e-6

    photons_in = np.zeros_like(wl_in)

    for c in wlc:
        photons_in +=  4e4 * np.exp(-0.5*((wl_in-c) / s)**2)

    photons_in += 4.0e4
    return photons_in


# A ThAR arc...
def create_th_ar_arc_spectrum(wl_in):

    lines = [(3719.41400, 58.52137),
             (3803.10800, 54.91138),
             (3828.37100, 56.63661),
             (3839.72400, 46.62416),
             (4019.13100, 40.06343),
             (4071.99600, 42.12776),
             (4131.76200, 50.57991),
             (4158.58400, 68.03227),
             (4200.65300, 48.00637),
             (4277.55800, 78.81951),
             (4348.11900, 80.24873)
             ]

    wl_in = wl_in.value
    s = 8e-6
    flux_out = np.zeros_like(wl_in)

    for cwl, flux in lines:
        c = cwl / 1e4
        flux_out +=  flux * np.exp(-0.5*((wl_in-c) / s)**2)

    return flux_out




from ..overplot_traces import main


def test_overplot_traces_help(capsys):
    """Check that the program runs with --help"""
    try:
        main(['--help'])
    except SystemExit as u:
        pass

    out, err = capsys.readouterr()
    assert True



#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy
import astropy.io.fits as fits
import uuid


def generate_multi_rss(imgs):

    nimages = len(imgs)
    if nimages == 0:
        raise ValueError('need at least 1 image')
    refimg = imgs[0]

    dim0 = refimg[0].shape[0]
    multishape = (dim0 * nimages, refimg[0].shape[1])
    fdata = numpy.empty(multishape)

    for idx, img in enumerate(imgs):
        fdata[idx*dim0: (idx+1)*dim0] = img[0].data

    hdu = fits.PrimaryHDU(data=fdata, header=refimg[0].header)
    hdu.header['MEG-NRSS'] = nimages
    # Generate a new UUID
    hdu.header['UUID'] = str(uuid.uuid1())

    allhdus = [hdu]
    for idx, img in enumerate(imgs, 1):
        fibers = img['FIBERS']
        fibers.header['EXTNAME'] = 'FIBERS{}'.format(idx)
        allhdus.append(fibers)
    result = fits.HDUList(allhdus)
    return result



import pytest

import numpy
import astropy.wcs

from ..fluxcalib import update_flux_limits, WAVLIM_KEYS, PIXLIM_KEYS


def generate_wcs():
    wcsl = astropy.wcs.WCS(naxis=2)

    crpix = 3.0
    crval = 5300.0
    cdelt = 0.1

    wcsl.wcs.crpix = [crpix, 0.0]
    wcsl.wcs.crval = [crval, 0.0]
    wcsl.wcs.cdelt = [cdelt, 1.0]
    return wcsl


@pytest.mark.parametrize("wcsl", [None, generate_wcs()])
@pytest.mark.parametrize("ref", [0, 1])
def test_update_flux_limits2(wcsl, ref):

    header = {}

    values = numpy.array([1, 1000, 2, 900, 100, 800])

    off = (ref + 1) % 2

    if wcsl is None:
        results = (values + off)
    else:
        lm = numpy.array([values, numpy.zeros_like(values)])
        wavelen_ = wcsl.all_pix2world(lm.T, ref)
        results = wavelen_[:, 0]

    fluxlimits = dict(zip(PIXLIM_KEYS, values))
    hdr = update_flux_limits(header, fluxlimits, wcs=wcsl, ref=ref)

    for key in WAVLIM_KEYS:
        assert key in hdr

    for key in PIXLIM_KEYS:
        assert key in hdr

    for key, val in zip(PIXLIM_KEYS, values):
        assert hdr[key][0] == val + off

    for key, res in zip(WAVLIM_KEYS, results):
        assert hdr[key][0] == res


def test_update_flux_invalid():

    header = {}

    values = [1, 1000, 2, 900, 100, 800]
    fluxlimits = dict(zip(PIXLIM_KEYS, values))

    ref = 2
    with pytest.raises(ValueError):
        update_flux_limits(header, fluxlimits, ref=ref)



#
# Copyright 2011-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Acquisition with LCB"""


import math

import numpy as np
from scipy.spatial import KDTree

from numina.core import Result, Parameter
from numina.core.validator import range_validator
from numina.constants import FWHM_G

from megaradrp.recipes.scientific.base import ImageRecipe
from megaradrp.types import ProcessedRSS, ProcessedFrame


class AcquireLCBRecipe(ImageRecipe):
    """Process Acquisition LCB images.

    This recipe processes a set of acquisition images
    obtained in **LCB Acquisition** mode and returns
    the offset and rotation required to center the
    fiducial object in its reference positions.

    See Also
    --------
    megaradrp.recipes.auxiliary.acquisitionmos.AcquireMOSRecipe

    Notes
    -----
    Images provided by `obresult` are trimmed and corrected
    from overscan, bad pixel mask (if `master_bpm` is not None),
    bias, dark current (if `master_dark` is not None) and
    slit-flat (if `master_slitflat` is not None).

    Images thus corrected are the stacked using the median.
    The result of the combination is saved as an intermediate result, named
    'reduced_image.fits'. This combined image is also returned in the field
    `reduced_image` of the recipe result.

    The apertures in the 2D image are extracted, using the information in
    `master_traces` and resampled according to the wavelength calibration in
    `master_wlcalib`. Then is divided by the `master_fiberflat`.
    The resulting RSS is saved as an intermediate
    result named 'reduced_rss.fits'. This RSS is also returned in the field
    `reduced_rss` of the recipe result.

    The sky is subtracted by combining the the fibers marked as `SKY`
    in the fibers configuration. The RSS with sky subtracted is returned ini the
    field `final_rss` of the recipe result.

    Then, the centroid of the fiducial object nearest to the center of the field
    is computed. The offset needed to center
    the fiducial object in the center of the LCB is returned.

    """

    # Requirements are defined in base class
    points = Parameter([(0, 0)], "Coordinates")
    nrings = Parameter(3, "Number of rings to extract the star",
                       validator=range_validator(minval=1))
    extraction_region = Parameter(
        [1000, 3000],
        description='Region used to compute a mean flux',
        nelem=2
    )


    reduced_image = Result(ProcessedFrame)
    reduced_rss = Result(ProcessedRSS)
    final_rss = Result(ProcessedRSS)
    offset = Result(list)
    rotang = Result(float)

    def run(self, rinput):

        self.logger.info('starting AC LCB reduction')

        reduced2d, reduced1d = super(AcquireLCBRecipe, self).base_run(rinput)
        # rssdata = rss_data[0].data

        do_sky_subtraction = True
        if do_sky_subtraction:
            self.logger.info('start sky subtraction')
            isb = rinput.ignored_sky_bundles
            if isb:
                self.logger.info('sky bundles ignored: %s', isb)
            final, origin, sky = self.run_sky_subtraction(reduced1d,
                                                          ignored_sky_bundles=isb)
            self.logger.info('end sky subtraction')
        else:
            final =  reduced1d
            origin = final
            sky = final

        fiberconf = self.datamodel.get_fiberconf(final)
        self.logger.debug("LCB configuration is %s", fiberconf.conf_id)

        rssdata = final[0].data

        scale, funit = self.datamodel.fiber_scale_unit(final, unit=True)
        self.logger.debug('unit is %s', funit)
        platescale = self.datamodel.PLATESCALE

        cut1, cut2 = rinput.extraction_region

        # points = [(0, 0)] # Center of fiber 313
        points = list(rinput.points)

        flux_per_cell_all = rssdata[:, cut1:cut2].mean(axis=1)

        max_cell = flux_per_cell_all.argmax() + 1
        max_fiber_ = fiberconf.fibers[max_cell]

        self.logger.info("maximum flux in spaxel %d -- %s", max_cell, max_fiber_.name)
        # Extend points with the brightest spaxel
        points.append((max_fiber_.x, max_fiber_.y))

        fibers = fiberconf.conected_fibers(valid_only=True)

        grid_coords = []
        for fiber in fibers:
            grid_coords.append((fiber.x, fiber.y))
        # setup kdtree for searching
        kdtree = KDTree(grid_coords)

        # Other posibility is
        # query using radius instead
        # radius = 1.2
        # kdtree.query_ball_point(points, k=7, r=radius)
        self.logger.debug('adding %d nrings', rinput.nrings)
        npoints = 1 + 3 * rinput.nrings * (rinput.nrings + 1)
        self.logger.debug('adding %d fibers', npoints)

        dis_p, idx_p = kdtree.query(points, k=npoints)

        self.logger.info('Using %d nearest fibers', npoints)
        for diss, idxs, point in zip(dis_p, idx_p, points):
            # For each point
            value = [p * scale for p in point]
            value_mm = [(v / platescale) for v in value]
            self.logger.info('For point %s arcsec', value)
            self.logger.info('For point %s mm', value_mm)
            colids = []
            coords = []
            for dis, idx in zip(diss, idxs):
                fiber = fibers[idx]
                colids.append(fiber.fibid - 1)
                coords.append((fiber.x, fiber.y))
            self.logger.debug('nearest fibers')
            self.logger.debug('%s', [col + 1 for col in colids])
            coords = np.asarray(coords) * scale
            # flux_per_cell = flux_per_cell_all[colids]
            flux_per_cell = rssdata[colids, cut1:cut2].mean(axis=1)
            flux_per_cell_total = flux_per_cell.sum()
            flux_per_cell_norm = flux_per_cell / flux_per_cell_total
            # centroid
            scf = coords.T * flux_per_cell_norm
            centroid = scf.sum(axis=1)
            self.logger.info('centroid: %s arcsec', list(centroid))
            self.logger.info('centroid: %s mm', list(centroid / platescale))
            # central coords
            c_coords = coords - centroid
            scf0 = scf - centroid[:, np.newaxis] * flux_per_cell_norm
            mc2 = np.dot(scf0, c_coords)
            self.logger.info('2nd order moments, x2=%f, y2=%f, xy=%f arcsec^2', mc2[0,0], mc2[1,1], mc2[0,1])
            self.logger.info('FWHM , x=%f, y=%f arcsec',
                             FWHM_G * math.sqrt(mc2[0,0]),
                             FWHM_G * math.sqrt(mc2[1,1])
                             )

        if False:
            self.compute_dar(final)

        return self.create_result(
            reduced_image=reduced2d,
            reduced_rss=reduced1d,
            final_rss=final,
            offset=-centroid
        )



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy as np
from astropy import units as u
from astropy.modeling.blackbody import blackbody_lambda
from numina.instrument.hwdevice import HWDevice

from megaradrp.simulation.extended import create_th_ar_arc_spectrum


class Lamp(HWDevice):
    def __init__(self, name, factor=1.0, illumination=None):
        super(Lamp, self).__init__(name=name)
        self.factor = factor
        if illumination is None:
            self._illum = lambda x, y: np.ones_like(x)
        else:
            self._illum = illumination

    def flux(self, wl):
        units = u.erg * u.s**-1 * u.cm ** -2 * u.AA**-1 * u.sr**-1
        return self.factor * np.ones_like(wl).value * units

    def illumination(self, x, y):
        return self._illum(x, y)


class BlackBodyLamp(Lamp):

    def __init__(self, name, temp, factor=1.0, illumination=None):
        self.temp = temp
        super(BlackBodyLamp, self).__init__(name, factor=factor,
                                            illumination=illumination)

    def flux(self, wl_in):
        energy_in_flat = blackbody_lambda(wl_in, self.temp)
        return self.factor * energy_in_flat


class FlatLamp(Lamp):
    def __init__(self, name, factor=1.0, illumination=None):
        super(FlatLamp, self).__init__(name, factor=factor,
                                       illumination=illumination)


class ArcLamp(Lamp):

    def flux(self, wl_in):
        val = create_th_ar_arc_spectrum(wl_in)
        val_u = val * u.erg * u.s**-1 * u.cm ** -2 * u.AA**-1 * u.sr**-1
        return self.factor * val_u



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import math

import numpy
from numina.instrument.hwdevice import HWDevice

from megaradrp.datamodel import TargetType


class RoboticPositioner(HWDevice):

    def __init__(self, name, id, pos=None, parent=None):
        super(RoboticPositioner,self). __init__(name, parent)
        self._id = id
        if pos is None:
            pos = (0.0, 0.0, 0.0)
        self.x_fix = pos[0]
        self.y_fix = pos[1]
        self.pa_fix = pos[2]

        self.x_ = self.x_fix
        self.y_ = self.y_fix
        self.pa_ = self.pa_fix

        self._target_priority = 0
        self._target_type = TargetType.UNASSIGNED
        self._target_name = 'unknown'

        self.fb = None
        self.rpatrol = 10.0

    def move_to(self, x, y, pa):
        dis  = math.hypot(x - self.x_fix, y - self.y_fix)
        if dis >= self.rpatrol:
            # impossible to move
            raise ValueError('movement out of patrol area')
        self.x_ = x
        self.y_ = y
        self.pa_ = pa

    def park(self):
        self.move_to(self.x_fix, self.y_fix, self.pa_fix)

    def connect_bundle(self, bundle):
        self.fb = bundle

    def fibers_in_focal_plane(self):
        # Given PA and center, return position
        # of the fibers

        base = math.pi / 3.0
        ini = base / 2.0
        rad = 0.5365 # Geometry
        PA = self.pa * (math.pi / 180)

        if self.fb is None:
            return ([], [])
        else:
            # To the left
            #angs = PA + ini + base * np.arange(6)
            # To the right
            angs = -PA + ini + base * numpy.arange(6)
            m0 = rad * numpy.cos(angs)
            m1 = rad * numpy.sin(angs)

            # Rearrange, fibers are counted in different order
            xx = m0[[2, 3, 1, 0, 4, 5, 0]]
            yy = m1[[2, 3, 1, 0, 4, 5, 0]]
            # This is the central fiber
            xx[3] = 0.0 # Center
            yy[3] = 0.0 # Center
            xx += self.x
            yy += self.y
            return self.fb.fibs_id, list(zip(xx, yy))

    @property
    def x(self):
        return self.x_

    @property
    def y(self):
        return self.y_

    @property
    def pa(self):
        return self.pa_

    @property
    def position(self):
        return self.x_, self.y_, self.pa_

    @position.setter
    def position(self, value):
        self.move_to(value[0], value[1], value[2])

    @property
    def position_relative(self):
        return self.x_ - self.x_fix, self.y_ - self.y_fix, self.pa_

    @position_relative.setter
    def position_relative(self, value):
        self.move_to(value[0] + self.x_, value[1] + self.y_, value[2] + self.pa_)

    @property
    def target_priority(self):
        return self._target_priority

    @target_priority.setter
    def target_priority(self, value):
        self._target_priority = value

    @property
    def target_type(self):
        return self._target_type

    @target_type.setter
    def target_type(self, value):
        self._target_type = value

    @property
    def target_name(self):
        return self._target_name

    @target_name.setter
    def target_name(self, value):
        self._target_name = value

    def init_config_info(self):
        meta = super(RoboticPositioner, self).init_config_info()
        meta["id"] = self._id
        if self.fb:
            meta['bundle'] = self.fb.config_info()
            _, pos = self.fibers_in_focal_plane()
            meta['fibers_pos'] = pos
        return meta


class BaseFibersPlane(HWDevice):
    def __init__(self, name, fiberset, parent=None):

        super(BaseFibersPlane, self).__init__(name, parent)
        self.fiberset = fiberset
        self.confid_ = 1

    @property
    def nbundles(self):
        return self.nfibers // 7

    @property
    def nfibers(self):
        return len(self.fiberset.fibers)

    @property
    def size(self):
        return self.fiberset.size

    @property
    def sigma(self):
        return self.fiberset.sigma

    @property
    def area(self):
        return self.fiberset.area

    @property
    def conf_id(self):
        return self.confid_

    @conf_id.setter
    def conf_id(self, value):
        self.confid_ = value

    def transmission(self, wlin):
        return self.fiberset.transmission(wlin)


class FiberMOS(BaseFibersPlane):
    def __init__(self, name, fiberset, parent=None):
        super(FiberMOS, self).__init__(name, fiberset, parent)

    def fibers_in_focal_plane(self):
        fibid = []
        pos = []
        for robot in self.children.values():
            # Compute positions of the fibers
            # in focal plane for fibers
            res1, res2 = robot.fibers_in_focal_plane()
            fibid.extend(res1)
            pos.extend(res2)
        # This must be sorted
        fibid = numpy.array(fibid)
        pos = numpy.array(pos)
        idx = numpy.argsort(fibid)
        return fibid[idx], pos[idx]

    def robots_in_focal_plane(self):
        robid = []
        pos = []
        for robot in self.children.values():
            # Compute positions of the fibers
            # in focal plane for fibers
            res2 = robot.position
            robid.append(robot._id)
            pos.append(res2[:2])
        return robid, pos


class LargeCompactBundle(BaseFibersPlane):
    def __init__(self, name, fiberset, lcb_pos, parent=None):
        super(LargeCompactBundle, self).__init__(name, fiberset, parent)
        self.lcb_pos = lcb_pos

    def fibers_in_focal_plane(self):
        fibid = [fiber.fibid for fiber in self.fiberset.fibers.values()]
        pos = [self.lcb_pos[fiber.fibid] for fiber in self.fiberset.fibers.values()]

        return fibid, pos

    @property
    def fibers(self):
        res = []
        for bundid in self.fiberset.bundles:
            for lf in self.fiberset.bundles[bundid].lf:
                pos = self.lcb_pos[lf.fibid]
                res.append((lf.fibid, bundid, pos[0], pos[1], lf.inactive))

        return res



#
# Copyright 2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Products of the Megara Pipeline"""

import itertools

import numpy


def to_ds9_reg(obj, ds9reg, rawimage=False, numpix=100, fibid_at=0):
    """Transform fiber traces to ds9-region format.

    Parameters
    ----------
    ds9reg : BinaryIO
        Handle to output file name in ds9-region format.
    rawimage : bool
        If True the traces must be generated to be overplotted on
        raw FITS images.
    numpix : int
        Number of abscissae per fiber trace.
    fibid_at : int
        Abscissae where the fibid is shown (default=0 -> not shown).

    """

    # offset between polynomial and image abscissae
    if rawimage:
        ix_offset = 51
    else:
        ix_offset = 1

    # open output file and insert header

    ds9reg.write('# Region file format: DS9 version 4.1\n')
    ds9reg.write(
        'global color=green dashlist=2 4 width=2 font="helvetica 10 '
        'normal roman" select=1 highlite=1 dash=1 fixed=0 edit=1 '
        'move=1 delete=1 include=1 source=1\n')
    ds9reg.write('physical\n')
    ds9reg.write('#\n# insmode: {0}\n'.format(obj.tags['insmode']))
    ds9reg.write('# vph: {0}\n'.format(obj.tags['vph']))
    ds9reg.write('# uuid: {0}\n'.format(obj.uuid))
    colorbox = ['#ff77ff', '#4444ff']
    itercolors = itertools.cycle(colorbox)
    for fiberdict in obj.contents:
        fibid = fiberdict.fibid
        # boxid = fiberdict.boxid
        xmin = fiberdict.start
        xmax = fiberdict.stop
        ds9reg.write('#\n# fibid: {0}\n'.format(fibid))
        # skip fibers without trace
        if fiberdict.valid:
            xp = numpy.linspace(start=xmin, stop=xmax, num=numpix)
            yp = fiberdict.polynomial(xp)
            if rawimage:
                lcut = (yp > 2056.5)
                yp[lcut] += 100
            for i in range(len(xp) - 1):
                x1 = xp[i] + ix_offset
                y1 = yp[i] + 1
                x2 = xp[i + 1] + ix_offset
                y2 = yp[i + 1] + 1
                ds9reg.write('line {0} {1} {2} {3}'.format(x1, y1, x2, y2))
                # Alternate colors
                next_color = next(itercolors)
                ds9reg.write(' # color={0}\n'.format(next_color))
                if fibid_at != 0:
                    if x1 <= fibid_at <= x2:
                        ds9reg.write('text {0} {1} {{{2}}} # color=green '
                                     'font="helvetica 10 bold '
                                     'roman"\n'.format((x1+x2)/2,
                                                       (y1+y2)/2, fibid))



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Corrector for wavelength calibration"""

import logging
import datetime
import warnings
import uuid

import numpy
from numpy.polynomial.polynomial import polyval
import astropy.wcs
import astropy.io.fits as fits

import numina.array.utils as utils
from numina.frame.utils import copy_img
from numina.processing import Corrector
from numina.array.interpolation import SteffenInterpolator

from megaradrp.instrument import WLCALIB_PARAMS

_logger = logging.getLogger(__name__)


class SimpleWcs1D(object):
    """Store parameters of a simple 1D WCS"""
    def __init__(self, crval=0, crpix=0, cdelt=1, ctype='', cunit='', **kwargs):
        self.crval = crval
        self.crpix = crpix
        self.cdelt = cdelt
        self.ctype = ctype
        self.cunit = cunit

    def create_internal_wcs_(self):
        # Not using units, only when writing in the header
        w = astropy.wcs.WCS(naxis=2, fix=False)
        w.wcs.crpix = [self.crpix, 0.0]
        w.wcs.cdelt = [self.cdelt, 1.0]
        w.wcs.crval = [self.crval, 0.0]
        w.wcs.ctype = ['AWAV', ' ']
        return w


def get_imgid(img, prefix=True):
    hdr = img[0].header
    base = "{}"
    if 'UUID' in hdr:
        pre = 'uuid:{}'
        value = hdr['UUID']
    elif 'DATE-OBS' in hdr:
        pre = 'dateobs:{}'
        value = hdr['DATE-OBS']
    elif 'checksum' in hdr:
        pre = 'checksum:{}'
        value = hdr['checksum']
    elif 'filename' in hdr:
        pre = 'file:{}'
        value = hdr['filename']
    else:
        raise ValueError('no method to identity image')
    if prefix:
        return pre.format(value)
    else:
        return base.format(value)


class WavelengthCalibrator(Corrector):
    """A Node that applies wavelength calibration."""

    def __init__(self, solutionwl, datamodel=None, dtype='float32'):

        super(WavelengthCalibrator, self).__init__(
            datamodel=datamodel,
            calibid=solutionwl.calibid,
            dtype=dtype)

        self.solutionwl = solutionwl

    def run(self, rss):

        newrss = calibrate_wl_rss_megara(
            rss, self.solutionwl,
            dtype=self.dtype, span=2, inplace=True
        )
        return newrss


def calibrate_wl_rss_megara(rss, solutionwl, dtype='float32', span=0, inplace=False):
    """Apply wavelength calibration to a RSS

    Parameters
    ----------

    rss: astropy.io.fits.HDUList
        A Row stacked Spectra MEGARA image, not WL calibrated
    solutionwl: megaradrp.products.wavecalibration.WavelengthCalibration
        A wavelength calibration solution
    dtype: str
        Convertible to numpy.dtype
    span: int
        Remove `span` pixels at both sides of the resampled image
    inplace: bool
        The input RSS is modified inplace, if False, a copy is made

    Returns
    -------
    rss: astropy.io.fits.HDUList
        A Row stacked Spectra MEGARA image, WL calibrated
    """

    imgid = get_imgid(rss)
    _logger.debug('wavelength calibration in image %s', imgid)
    _logger.debug('with wavecalib %s', solutionwl.calibid)
    _logger.debug('offsets are %s', solutionwl.global_offset.coef)

    current_vph = rss[0].header['VPH']
    current_insmode = rss[0].header['INSMODE']

    _logger.debug('Current INSMODE is %s, VPH is %s', current_insmode, current_vph)
    if current_insmode in WLCALIB_PARAMS and current_vph in WLCALIB_PARAMS[current_insmode]:
        wvpar_dict = WLCALIB_PARAMS[current_insmode][current_vph]
        _logger.info('precomputed wl parameters are %s', wvpar_dict)
    else:
        msg = 'insmode {} grism {} is not defined in megaradrp.instrument.WLCALIB_PARAMS'.format(
            current_insmode,
            current_vph
        )
        raise ValueError(msg)

    targetwcs = SimpleWcs1D(**wvpar_dict)
    npix = wvpar_dict['npix']

    result = calibrate_wl_rss(
        rss, solutionwl, npix, targetwcs,
        dtype=dtype,
        span=span, inplace=inplace
    )
    return result


def calibrate_wl_rss(rss, solutionwl, npix, targetwcs, dtype='float32', span=0, inplace=False):
    """Apply wavelength calibration to a RSS

    Parameters
    ----------

    rss: astropy.io.fits.HDUList
        A Row stacked Spectra MEGARA image, not WL calibrated
    solutionwl: megaradrp.products.wavecalibration.WavelengthCalibration
        A wavelength calibration solution
    npix: int
        Number of channels of the calibrated RSS
    targetwcs: SimpleWcs1D
        Common WCS solution
    dtype: str
        Convertible to numpy.dtype
    span: int
        Remove `span` pixels at both sides of the resampled image
    inplace: bool
        The input RSS is modified inplace, if False, a copy is made

    Returns
    -------
    rss: astropy.io.fits.HDUList
        A Row stacked Spectra MEGARA image, WL calibrated
    """

    if not inplace:
        # This is a new HDUList
        rss = copy_img(rss)

    imgid = get_imgid(rss)
    _logger.debug('wavelength calibration in image %s', imgid)
    _logger.debug('with wavecalib %s', solutionwl.calibid)
    _logger.debug('offsets are %s', solutionwl.global_offset.coef)

    # Target WCS
    re_wcs = targetwcs.create_internal_wcs_()

    _logger.debug('Resample RSS')
    final, limits = resample_rss_flux(
        rss[0].data, solutionwl, npix, re_wcs,
        span=span, fill=0
    )

    rss[0].data = final.astype(dtype)

    hdr = rss[0].header
    _logger.debug('Add WCS headers')
    rss_add_wcs(hdr, targetwcs.crval, targetwcs.cdelt, targetwcs.crpix)

    _logger.debug('Add calibration headers')
    hdr['NUM-WAV'] = solutionwl.calibid
    hdr['history'] = 'Wavelength calibration with {}'.format(solutionwl.calibid)
    hdr['history'] = 'Aperture extraction offsets are {}'.format(
        solutionwl.global_offset.coef.tolist())
    hdr['history'] = 'Wavelength calibration time {}'.format(datetime.datetime.utcnow().isoformat())
    hdr['history'] = 'Resample span={}'.format(span)
    # Update UUID
    hdr['UUID'] = str(uuid.uuid1())

    # Update other HDUs if needed
    # dtype here can be int16 or uint8
    map_data = numpy.zeros_like(final, dtype='int16')

    fibers_ext = rss['FIBERS']
    fibers_ext_headers = fibers_ext.header
    # Add KEYWORDS
    # FIB%03dW1, FIB%03dW2
    for fibid, (lower, upper) in limits:
        idx = fibid - 1
        map_data[idx, lower:upper+1] = 1
        # Update Fibers
        key = "FIB{:03d}W1".format(fibid)
        fibers_ext_headers[key] =  (lower + 1, "Start of spectral coverage")
        key = "FIB{:03d}W2".format(fibid)
        fibers_ext_headers[key] =  (upper + 1, "End of spectral coverage")

    # Update KEYWORDS
    # "FIB%03d_V"
    for fibid in solutionwl.error_fitting:
        # Update Fibers
        key = "FIB%03d_V" % fibid
        fibers_ext_headers[key] =  False

    for fibid in solutionwl.missing_fibers:
        # Update Fibers
        key = "FIB%03d_V" % fibid
        fibers_ext_headers[key] =  False

    rss_map = fits.ImageHDU(data=map_data, name='WLMAP')

    rss.append(rss_map)
    return rss


def rss_add_wcs(hdr, crval, cdelt, crpix):
    """Add MEGARA 2D wavelength calibration headers"""
    c_crpix = 'Pixel coordinate of reference point'
    c_cunit = 'Units of coordinate increment and value'
    unit = 'Angstrom'
    c_crval = 'Coordinate value at reference point'
    c_cdelt = 'Coordinate increment at reference point'
    hdr['CRPIX1'] = crpix, c_crpix
    hdr['CRVAL1'] = crval, c_crval
    hdr['CDELT1'] = cdelt, c_cdelt
    hdr['CUNIT1'] = unit, c_cunit
    hdr['CTYPE1'] = 'AWAV', 'Air wavelength (linear)'

    hdr['CRPIX2'] = 0.0, c_crpix
    hdr['CRVAL2'] = 0.0, c_crval
    hdr['CDELT2'] = 1.0, c_cdelt
    hdr['CTYPE2'] = ''
    return hdr


def create_internal_wcs_(wlr0, delt, crpix):
    # Not using units, only when writing in the header
    w = astropy.wcs.WCS(naxis=2, fix=False)
    w.wcs.crpix = [crpix, 0.0]
    w.wcs.cdelt = [delt, 1.0]
    w.wcs.crval = [wlr0, 0.0]
    w.wcs.ctype = ['AWAV', ' ']

    return w


def resample_rss_flux(arr, solutionwl, npix, finalwcs, span=0, fill=0):
    """Resample array according to a wavelength calibration solution

    Parameters
    ----------

    arr: numpy.ndarray
        A Row stacked Spectra MEGARA image, not WL calibrated
    solutionwl: megaradrp.products.wavecalibration.WavelengthCalibration
        A wavelength calibration solution
    npix: int
        Number of channels of the calibrated RSS
    finalwcs: astropy.wcs.WCS
        WCS solution of the final array
    span: int
        Remove `span` pixels at both sides of the resampled image
    fill: int
        Value used to fill the values removed by `span`

    Returns
    -------
    rss_resampled:  numpy.ndarray
        A resample array
    limits: a list of tuples
        Contains the fiberid and a pair with the first and last valid pixel (0-based)
    """

    nfibers = arr.shape[0]
    nsamples = arr.shape[1]

    # Use only the spectral axis
    subwcs = finalwcs.sub(['spectral'])

    # wl_max,  = subwcs.all_pix2world([npix], 1)

    # 0-based index
    new_x = numpy.arange(npix)
    new_wl,  = subwcs.all_pix2world(new_x, 0)

    # 0-based left borders
    old_x_borders_0 = numpy.arange(-0.5, nsamples)
    # 1-based left borders
    old_x_borders_1 = old_x_borders_0 + 1.0  # following FITS criterium

    # In AA
    new_wl_borders = pixel_borders(new_wl)

    accum_flux = numpy.empty((nfibers, nsamples + 1))
    accum_flux[:, 1:] = numpy.cumsum(arr, axis=1)
    accum_flux[:, 0] = 0.0
    rss_resampled = numpy.zeros((nfibers, npix))
    limits = []

    for fibsol in solutionwl.contents:

        fibid = fibsol.fibid
        idx = fibid - 1

        # small correction defined in master_wlcalib_XXX_XX-X.json
        coeff = fibsol.solution.coeff
        offset_wavelength = solutionwl.global_offset(fibid)
        coeff[0] -= offset_wavelength

        # Polynomial returns AA
        old_wl_borders = polyval(old_x_borders_1, coeff)
        # 0-based, AA
        ss_vals, = subwcs.all_world2pix(old_wl_borders[[0, -1]], 0)
        # s1 is the 0-based pixel that contains the lower limit
        # s2 is the 0-based pixel that contains the upper limit
        s1, s2 = ss_vals
        s1 = utils.coor_to_pix_1d(s1)
        s2 = utils.coor_to_pix_1d(s2)
        lower = max(0, min(s1, npix - 1))
        upper = max(0, min(s2, npix - 1))

        if lower > upper:
            warnings.warn('lower limit is > upper limit', RuntimeWarning)

        # We need a monotonic interpolator
        # linear would work, we use a cubic interpolator
        interpolator = SteffenInterpolator(
            old_wl_borders,
            accum_flux[idx],
            extrapolate='border'
        )

        if lower + span > upper - span:
            warnings.warn('lower limit + span is > upper limit - span', RuntimeWarning)

        fl_borders = interpolator(new_wl_borders)
        rss_resampled[idx] = fl_borders[1:] - fl_borders[:-1]
        # Expand the border to remove `span` pixels
        # in both sides, to avoid high variance
        rss_resampled[idx, lower:lower + span] = fill
        rss_resampled[idx, upper + 1 - span:upper + 1] = fill
        limits.append((fibid, (lower + span, upper - span)))

    return rss_resampled, limits


def pixel_borders(arr):
    import numina.array.wavecalib.resample as W
    return W.map_borders(arr)


if __name__ == '__main__':
    import numina.types.structured as stru
    import logging

    logging.basicConfig(level=logging.DEBUG)

    rssname = "/home/spr/Documentos/Congresos/ria_megara/M15_LCB_HR-R/obsid3_results/reduced_rss.fits"
    rss = fits.open(rssname)
    wlname = "/home/spr/Documentos/Congresos/ria_megara/M15_LCB_HR-R/obsid2_results/master_wlcalib.json"
    solutionwl = stru.open(wlname)

    rss2 = calibrate_wl_rss_megara(rss, solutionwl)
    rss2.writeto('result_span0b.fits', overwrite=True)

    rss2 = calibrate_wl_rss_megara(rss, solutionwl, span=2)
    rss2.writeto('result_span2b.fits', overwrite=True)


#
# Copyright 2014-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from numina.core import Result
from numina.array import combine

from megaradrp.processing.combine import basic_processing_with_combination
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.types import MasterBias
from megaradrp.requirements import MasterBPMRequirement


class BiasRecipe(MegaraBaseRecipe):
    """Process BIAS images and create a MASTER_BIAS product.

    This recipe process a set of bias images obtained in
    **Bias Image** mode and returns a combined product image,
    trimmed to the physical size of the detector.

    Notes
    -----
    Images are corrected from overscan and trimmed to the physical size of the detector.
    Then, they corrected from Bad Pixel Mask, if the BPM is available,
    Finally, images are stacked using the median.

    See Also
    --------
    megaradrp.types.MasterBias: description of the MasterBias product

    """

    master_bpm = MasterBPMRequirement()
    master_bias = Result(MasterBias)

    def run(self, rinput):
        """Execute the recipe.

        Parameters
        ----------

        rinput : BiasRecipe.RecipeInput

        Returns
        -------
        BiasRecipe.RecipeResult

        """
        self.logger.info('start bias recipe')
        flow = self.init_filters(rinput, rinput.obresult.configuration)
        errors  = False
        if not errors:
            self.logger.info('not computing errors')
        hdulist = basic_processing_with_combination(rinput, flow, method=combine.median, errors=errors)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        result = self.create_result(master_bias=hdulist)
        self.logger.info('end bias recipe')
        return result

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(BiasRecipe, self).set_base_headers(hdr)
        hdr['NUMTYPE'] = ('MasterBias', 'Product type')
        return hdr



# from megaradrp.processing.aperture import ApertureExtractor
# from numina.core import DataFrame
# import shutil
# from tempfile import mkdtemp
# import numpy as np
# import astropy.io.fits as fits
#
# def test_Aperture_Extractor():
#
#     eq = 0.8 * np.ones((4112, 4096))
#     temporary_path = mkdtemp()
#
#     fits.writeto('%s/eq.fits' % temporary_path, eq, overwrite=True)
#
#     image =file(temporary_path + '/eq.fits')
#     obj = ApertureExtractor(None)
#     # result = obj._run(image)
#     # print (result)
#
#
# if __name__ == "__main__":
#     test_Aperture_Extractor()



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Focus Spectrograph Recipe for Megara"""

from __future__ import division, print_function

import numpy
import numpy.polynomial.polynomial as polynomial
from scipy.spatial import cKDTree
import astropy.io.fits as fits

from numina.array import combine
from numina.core import Requirement, Parameter
from numina.core.dataholders import Result
from numina.types.array import ArrayType
from numina.core.requirements import ObservationResultRequirement
from numina.exceptions import RecipeError
import numina.array.utils
import numina.array.fwhm as fmod
import numina.core.validator
from numina.array.stats import robust_std as sigmaG
from numina.array.peaks.peakdet import find_peaks_indexes, refine_peaks

from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.types import FocusWavelength, ProcessedFrame
import megaradrp.requirements as reqs
from megaradrp.processing.combine import basic_processing_with_combination_frames
from megaradrp.processing.aperture import ApertureExtractor
from megaradrp.instrument import vph_thr_arc


class FocusSpectrographRecipe(MegaraBaseRecipe):
    """Process spectrograph focus images and find best focus.

    This recipe process a set of focus images obtained in
    **Focus Spectrograph** mode and returns different
    measurements of the spectrograph focus along de detector.

    See Also
    --------
    megaradrp.recipes.auxiliary.focustel.FocusTelescopeRecipe:
                recipe to measure the focus of the telescope

    Notes
    -----
    Images provided in `obresult` are grouped by the value of their
    FOCUS keyword. Groups of images are trimmed and corrected from overscan,
    bad pixel mask (if `master_bpm` is not None), bias and dark current
    (if `master_dark` is not None). Each group is then stacked using the median.

    The result of the combination is saved as an intermediate result, named
    'focus2d-#focus.fits', with #focus being the value of the focus
    of each group. Apertures are extracted in each combined image, and the
    resulting RSS file is saved as an intermediate result, named
    'focus1-#focus.fits'.

    For each image, peaks are detected every `nfibers` fibers, and their
    position, peak flux and FWHM is computed. The image with median
    focus is taken as reference image, and the peaks of every other
    image are matched against it.

    Then, for each line matched in the series of images, its FWHM
    is fitted to a 2nd degree polynomial, and the focus corresponding
    to its minimum is obtained.

    The recipe returns:

     * `focus_table`: a table with (x,y,best_focus) for each matched peak,
        with x,y measured in the reference bidimensional image

     * `focus_wavelength`: a structure containing measurements of every
        matched peak in each image

     * `focus_image`: a bidimensional image representing the spatial
        variation of the best focus, using a Voronoi diagram.

    """

    # Requirements
    obresult = ObservationResultRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_bpm = reqs.MasterBPMRequirement()
    master_traces = reqs.MasterAperturesRequirement()
    extraction_offset = Parameter([0.0], 'Offset traces for extraction', accept_scalar=True)
    master_wlcalib = reqs.WavelengthCalibrationRequirement()

    nfibers = Parameter(10, "The results are sampled every nfibers")
    tsigma = Parameter(50, "Scale factor for row threshold")
    # Results
    focus_table = Result(ArrayType)
    focus_image = Result(ProcessedFrame)
    focus_wavelength = Result(FocusWavelength)

    @numina.core.validator.validate
    def run(self, rinput):
        """Execute the recipe.

        Parameters
        ----------
        rinput : RecipeInput

        Returns
        -------
        RecipeResult

        Raises
        ------
        ValidationError
              if the recipe input is invalid

        """
        # Basic processing
        self.logger.info('start focus spectrograph')

        obresult = rinput.obresult

        flow = self.init_filters(rinput, obresult.configuration)

        current_vph = obresult.tags['vph']
        current_insmode = obresult.tags['insmode']

        if current_insmode in vph_thr_arc and current_vph in vph_thr_arc[current_insmode]:
            flux_limit = vph_thr_arc[current_insmode][current_vph].get('flux_limit', 200000)
            self.logger.info('flux_limit for %s is %4.2f', current_vph, flux_limit)
        else:
            flux_limit = 40000
            self.logger.info('flux limit not defined for %s, using %4.2f', current_vph, flux_limit)

        image_groups = {}
        self.logger.info('group images by focus')

        for idx, frame in enumerate(obresult.frames):
            with frame.open() as img:
                focus_val = img[0].header['focus']
                if focus_val not in image_groups:
                    self.logger.debug('new focus %s', focus_val)
                    image_groups[focus_val] = []
                self.logger.debug('image %s in group %s', img, focus_val)
                image_groups[focus_val].append(frame)

        if len(image_groups) < 2:
            raise RecipeError('We have only {} different focus'.format(len(image_groups)))

        # Loop only over fibers with WL calibration
        valid_traces = [fibsol.fibid for fibsol in rinput.master_wlcalib.contents]
        # valid_traces = [aper.fibid for aper in rinput.tracemap.contents if aper.valid]
        # every tenth fiber
        nfibers = rinput.nfibers
        valid_traces = valid_traces[::nfibers]

        ever = {}
        for focus, frames in image_groups.items():
            self.logger.info('processing focus %s', focus)

            try:
                img = basic_processing_with_combination_frames(frames, flow, method=combine.median, errors=False)
                calibrator_aper = ApertureExtractor(
                    rinput.master_traces,
                    self.datamodel,
                    offset=rinput.extraction_offset
                )

                self.save_intermediate_img(img, 'focus2d-%s.fits' % (focus,))
                img1d = calibrator_aper(img)
                self.save_intermediate_img(img1d, 'focus1d-%s.fits' % (focus,))

                self.logger.info('find lines and compute FWHM')
                lines_rss_fwhm = self.run_on_image(img1d, rinput.master_traces,
                                                   flux_limit,
                                                   valid_traces=valid_traces,
                                                   times_sigma=rinput.tsigma
                                                   )
                ever[focus] = lines_rss_fwhm

            except ValueError:
                self.logger.info('focus %s cannot be processed', focus)

        self.logger.info('pair lines in images')
        line_fibers = self.filter_lines(ever)

        focus_wavelength = self.generate_focus_wl(ever, rinput.master_wlcalib)

        self.logger.info('fit FWHM of lines')
        final = self.reorder_and_fit(line_fibers, sorted(image_groups.keys()))

        focus_median = numpy.median(final[:, 2])
        self.logger.info('median focus value is %5.2f', focus_median)

        self.logger.info('generate focus image')
        image = self.generate_image(final)
        focus_image_hdu = fits.PrimaryHDU(image)
        focus_image = fits.HDUList([focus_image_hdu])

        self.logger.info('end focus spectrograph')
        return self.create_result(focus_table=final, focus_image=focus_image,
                                  focus_wavelength=focus_wavelength)

    def run_on_image(self, img, tracemap, flux_limit=40000, valid_traces=None, times_sigma=50):
        """Extract spectra, find peaks and compute FWHM."""

        rssdata = img[0].data

        if valid_traces:
            valid_traces_s = set(valid_traces) # use set for fast membership
            valid_apers = [aper for aper in tracemap.contents if aper.valid and aper.fibid in valid_traces_s]
        else:
            valid_apers = [aper for aper in tracemap.contents if aper.valid]

        nwinwidth = 5
        lwidth = 20
        fpeaks = {}

        for aper in valid_apers:
            fibid = aper.fibid
            idx = fibid - 1
            row = rssdata[idx, :]

            the_pol = aper.polynomial

            # FIXME: using here a different peak routine than in arc
            # find peaks
            threshold = numpy.median(row) + times_sigma * sigmaG(row)
            self.logger.debug('values for threshold: median: %f, scale: %f, sigma: %f',
                              numpy.median(row), times_sigma, sigmaG(row))
            self.logger.debug('threshold is: %f', threshold)
            ipeaks_int1 = find_peaks_indexes(row, nwinwidth, threshold)
            # filter by flux
            self.logger.info('Filtering peaks over %5.0f', flux_limit)
            ipeaks_vals = row[ipeaks_int1]
            mask = ipeaks_vals < flux_limit
            ipeaks_int = ipeaks_int1[mask]
            self.logger.debug('LEN (ipeaks_int): %s', len(ipeaks_int))
            self.logger.debug('ipeaks_int: %s', ipeaks_int)
            ipeaks_float = refine_peaks(row, ipeaks_int, nwinwidth)[0]

            # self.pintarGrafica(refine_peaks(row, ipeaks_int, nwinwidth)[0] - refinePeaks_spectrum(row, ipeaks_int, nwinwidth))

            fpeaks[fibid] = []
            for peak, peak_f in zip(ipeaks_int, ipeaks_float):
                try:
                    sl = numina.array.utils.slice_create(peak, lwidth)
                    rel_peak = peak - sl.start
                    qslit = row[sl]
                    peak_val, fwhm = fmod.compute_fwhm_1d_simple(qslit, rel_peak)
                    peak_on_trace = the_pol(peak)
                    fpeaks[fibid].append((peak_f, peak_on_trace, fwhm))
                except ValueError as error:
                    self.logger.warning('Error %s computing FWHM in fiber %d', error, fibid)
                except IndexError as error:
                    self.logger.warning('Error %s computing FWHM in fiber %d', error, fibid)
            self.logger.debug('found %d peaks in fiber %d', len(fpeaks[fibid]), fibid)
        return fpeaks

    def generate_image(self, final):
        # FIXME: hardcoded sizes
        voronoi_points = numpy.array(final[:, [0, 1]])
        x = numpy.arange(2048 * 2)
        y = numpy.arange(2056 * 2)

        test_points = numpy.transpose(
            [numpy.tile(x, len(y)), numpy.repeat(y, len(x))])

        voronoi_kdtree = cKDTree(voronoi_points)

        test_point_dist, test_point_regions = voronoi_kdtree.query(test_points, k=1)
        final_image = test_point_regions.reshape((4112, 4096)).astype('float32')
        final_image[:, :] = final[final_image[:, :].astype('int32'), 2]
        return final_image

    def generate_focus_wl(self, all_measures, wlcalib):


        self.logger.info('start result generation')

        result = {}

        wlfib = {}
        for s in wlcalib.contents:
            wlfib[s.fibid] = s.solution

        for focus, image in all_measures.items():
            cresult = {}
            result[focus] = cresult
            for fiber, value in image.items():
                cresult[fiber] = []
                for arco in value:
                    try:
                        # FIXME: hardcoded sizes
                        x = 2048 * 2 - arco[0]
                        res = polynomial.polyval(x, wlfib[fiber].coeff)
                        cresult[fiber].append([arco[0], arco[1], arco[2], res])
                    except KeyError:
                        self.logger.warning("Fiber %d hasn't WL calibration, skipping", fiber)

        self.logger.info('end result generation')

        return result

    def pintarGrafica(self, diferencia_final):
        if False:
            import matplotlib.pyplot as plt

            fig = plt.figure(1)
            ax = fig.add_subplot(111)

            ejeX = numpy.arange(len(diferencia_final))
            ax.plot(ejeX, diferencia_final, label="0")
            # lgd = ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.1), ncol=4,
            #                 mode="expand")
            lgd = ax.legend(bbox_to_anchor=(0., 1.02, 1., .102),
                            loc='upper center', ncol=4, mode="expand",
                            borderaxespad=0.)
            handles, labels = ax.get_legend_handles_labels()

            fig.savefig('diferencia.eps', format='eps', dpi=1500,
                        bbox_extra_artists=(lgd,), bbox_inches='tight')
            plt.draw()
            plt.show()

    def filter_lines(self, all_measures, maxdis=2.0):
        """Match lines between different images """

        values = sorted(all_measures.keys())
        ntotal = len(values)
        center = ntotal // 2
        center_focus = values[center]
        base = all_measures[center_focus]
        self.logger.debug('use image #%d as reference, focus=%s', center, center_focus)
        line_fibers = {}

        self.logger.debug('matching lines up to %3.1f pixels', maxdis)
        for fiberid in base:
            self.logger.debug('Using %d fiber in reference image', fiberid)
            ref = numpy.array(base[fiberid])
            if ref.size == 0:
                self.logger.debug('Reference fiber has no lines, skip')
                continue

            self.logger.debug('Positions are %s', ref)
            outref = len(ref)
            savelines = {}
            line_fibers[fiberid] = savelines
            for i in range(outref):
                savelines[i] = {}
                savelines[i]['basedata'] = {}
                savelines[i]['centers'] = []
            self.logger.debug('Create kd-tree in fiber %d', fiberid)
            kdtree = cKDTree(ref[:, :2])
            for i in range(ntotal):
                if i == center:
                    for j in range(outref):
                        savelines[j]['basedata']['coordinates'] = tuple(
                            ref[j, :2])
                        savelines[j]['centers'].append(ref[j, 2])
                    continue
                self.logger.debug('Matching lines in fiber %d in image # %d', fiberid, i)
                comp = numpy.array(all_measures[values[i]][fiberid])

                if comp.size == 0:
                    self.logger.debug('No lines in fiber %d in image # %d', fiberid, i)
                    continue
                else:
                    self.logger.debug('Using %d lines in fiber %d in image # %d', comp.size, fiberid, i)

                qdis, qidx = kdtree.query(comp[:, :2],
                                          distance_upper_bound=maxdis)
                for compidx, lidx in enumerate(qidx):
                    if lidx < outref:
                        savelines[lidx]['centers'].append(comp[compidx, 2])

            remove_groups = []

            for ir in savelines:
                if len(savelines[ir]['centers']) != ntotal:
                    remove_groups.append(ir)

            for ir in remove_groups:
                self.logger.debug('remove group of lines %d in fiber %d', ir,
                              fiberid)
                del savelines[ir]

        return line_fibers

    def reorder_and_fit(self, line_fibers, focii):
        """Fit all the values of FWHM to a 2nd degree polynomial and return minimum."""

        l = sum(len(value) for key, value in line_fibers.items())
        self.logger.debug('there are %d groups of lines to fit', l)
        ally = numpy.zeros((len(focii), l))
        final = numpy.zeros((l, 3))
        self.logger.debug('focci are %s', focii)
        l = 0
        for i in line_fibers:
            for j in line_fibers[i]:
                ally[:, l] = line_fibers[i][j]['centers']
                final[l, :2] = line_fibers[i][j]['basedata']['coordinates']
                l += 1

        self.logger.debug('line widths are %s', ally)
        try:
            res = numpy.polyfit(focii, ally, deg=2)
            self.logger.debug('fitting to deg 2 polynomial, done')
            best = -res[1] / (2 * res[0])
            final[:, 2] = best
        except ValueError as error:
            self.logger.warning("Error in fitting: %s", error)
            final[:, 2] = 0.0

        return final



import shutil
from tempfile import mkdtemp
import os

import astropy.io.fits as fits
import numpy as np
import pytest

from megaradrp.processing.trimover import trimOut
from megaradrp.processing.trimover import apextract_weights
from megaradrp.instrument.components.detector import ReadParams, MegaraDetectorSat
from megaradrp.simulation.actions import simulate_flat


def generate_bias_file():
    PSCAN = 50
    DSHAPE = (2056 * 2, 2048 * 2)
    OSCAN = 50

    ron = 2.0
    gain = 1.0
    bias = 1000.0

    qe = 0.8 * np.ones(DSHAPE)
    qe[0:15, 0:170] = 0.0

    readpars1 = ReadParams(gain=gain, ron=ron, bias=bias)
    readpars2 = ReadParams(gain=gain, ron=ron, bias=bias)

    detector = MegaraDetectorSat('megara_test_detector', DSHAPE, OSCAN, PSCAN, qe=qe,
                                 dark=(3.0 / 3600.0),
                                 readpars1=readpars1, readpars2=readpars2,
                                 bins='11')

    return simulate_flat(detector, exposure=1.0, source=5000.0)


@pytest.mark.parametrize("direction", ['normal', 'mirror'])
def test_trim_and_o(direction):
    temporary_path = mkdtemp()
    detconf = {
        'trim1': [[0,2056],[50,4146]],
        'trim2': [[2156,4212],[50,4146]],
        'bng': [1,1]
    }
    fs = generate_bias_file()
    fits.writeto('%s/flat.fits' % (temporary_path), fs, overwrite=True)
    trimOut(
        '%s/flat.fits' % (temporary_path),
        detconf,
        out='%s/result.fits' % (temporary_path),
        direction=direction,
    )
    with fits.open('%s/result.fits' % (temporary_path)) as hdul:
        assert hdul[0].shape[0] + 100 == fs.shape[0]
        assert hdul[0].shape[1] + 100 == fs.shape[1]

    shutil.rmtree(temporary_path)


def test_trim_and_o_fail():
    temporary_path = mkdtemp()
    fs = generate_bias_file()
    fits.writeto('%s/flat.fits' % (temporary_path), fs, overwrite=True)

    direction = 'fails'
    detconf = {
        'trim1': [[0,2056],[50,4146]],
        'trim2': [[2156,4212],[50,4146]],
        'bng': [1,1]
    }
    with pytest.raises(ValueError) as excinfo:
        trimOut(
            '%s/flat.fits' % (temporary_path),
            detconf,
            out='%s/result.fits' % (temporary_path),
            direction=direction
        )
    shutil.rmtree(temporary_path)
    assert excinfo.value.args[0] == "%s must be either 'normal' or 'mirror'" % direction


def test_trim_and_o_fail2():
    temporary_path = mkdtemp()
    fs = generate_bias_file()
    fits.writeto('%s/flat.fits' % (temporary_path), fs, overwrite=True)

    bins = 'fail'
    detconf = {
        'trim1': [[0,2056],[50,4146]],
        'trim2': [[2156,4212],[50,4146]],
        'bng': bins
    }
    with pytest.raises(ValueError) as excinfo:
        trimOut(
            '%s/flat.fits' % (temporary_path),
            detconf,
            out='%s/result.fits' % (temporary_path)
        )
    shutil.rmtree(temporary_path)
    assert excinfo.value.args[0] == "%s must be one if '11', '12', '21, '22'" % bins


@pytest.mark.skipif(not os.path.exists('master_weights_LCB_10img_1exp.tar'),
                    reason="no way of currently testing this without tar file")
def test_apextract_weights():
    import tarfile

    file_name = 'master_weights_LCB_10img_1exp.tar'

    data = fits.getdata('fiberflat_frame.fits')
    rss = apextract_weights(data, tarfile.open(file_name, 'r'))
    hdu_rss = fits.PrimaryHDU(rss)
    final = fits.HDUList([hdu_rss])
    final.writeto('rss.fits', overwrite=True)
    assert True


if __name__ == "__main__":
    test_trim_and_o('normal')
    # test_trim_and_o_fail()
    # test_trim_and_o_fail2()
    # test_apextract_weights()


#
# Copyright 2011-2017 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""LCB Fast Mapping Recipe for Megara"""


from numina.core import Product, ObservationResult

from megaradrp.types import ProcessedMultiRSS
from megaradrp.core.recipe import MegaraBaseRecipe
from megaradrp.processing.multirss import generate_multi_rss


class LCBFastMappingRecipe(MegaraBaseRecipe):
    """Process LCB Fast Mapping Recipe.

    This recipe processes a set of images
    obtained in **LCB Fast Mapping image** mode and
    a multi RSS image.

    See Also
    --------
    megaradrp.recipes.scientific.lcb.LCBImageRecipe
    megaradrp.recipes.scientific.mos.MOSImageRecipe

    Notes
    -----
    Images previously obtained in **LCB image** and reduced
    are stacked together in multi RSS format.

    """
    final_multirss = Product(ProcessedMultiRSS)

    def run(self, rinput):
        self.logger.info('start FastMappingRecipe')
        obresult = rinput.obresult
        imgs = [frame.open() for frame in obresult.frames]

        result = generate_multi_rss(imgs)
        self.logger.info('end FastMappingRecipe')

        return self.create_result(final_multirss=result)

    @classmethod
    def build_recipe_input(cls, obsres, dal, pipeline='default'):
        """Quey previous results of LCB image"""
        return cls.build_recipe_input_gtc(obsres, dal, pipeline=pipeline)

    @classmethod
    def build_recipe_input_gtc(cls, obsres, dal, pipeline='default'):
        cls.logger.debug('start recipe input builder')
        stareImagesIds = obsres.stareImagesIds
        cls.logger.debug('LCB image IDS %s: ', stareImagesIds)
        stareImages = []
        for subresId in stareImagesIds:
            subres = dal.getRecipeResult(subresId)
            stareImages.append(subres['elements']['final_rss'])

        newOR = ObservationResult()
        newOR.frames = stareImages
        newRI = cls.create_input(obresult=newOR)
        cls.logger.debug('end recipe input builder')
        return newRI



#
# Copyright 2017-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import print_function

import math

import matplotlib.cbook as cbook
import matplotlib.collections as mcoll
import matplotlib.colors as mcolors
import matplotlib.transforms as mtrans
import matplotlib.transforms as mtransforms
import numpy as np


M_SQRT3 = math.sqrt(3)

PLATESCALE = 1.2120  # arcsec / mm
SCALE = 0.443  # mm from center to center, upwards


def hexplot(axis, x, y, z, scale=1.0, extent=None,
            cmap=None, norm=None, vmin=None, vmax=None,
            alpha=None, linewidths=None, edgecolors='none',
            **kwargs):
    """
    Make a hexagonal grid plot.

    Returns
    =======
    object: matplotlib.collections.PolyCollection

    """
    axis._process_unit_info(xdata=x, ydata=y, kwargs=kwargs)

    x, y, z = cbook.delete_masked_points(x, y, z)

    x = np.array(x, float)
    y = np.array(y, float)

    sx = 2 / M_SQRT3 * scale * 0.99
    sy = scale * 0.99

    if extent is not None:
        xmin, xmax, ymin, ymax = extent
    else:
        xmin, xmax = (np.amin(x - sx), np.amax(x + sx)) if len(x) else (0, 1)
        ymin, ymax = (np.amin(y - sy), np.amax(y + sy)) if len(y) else (0, 1)

        # to avoid issues with singular data, expand the min/max pairs
        xmin, xmax = mtrans.nonsingular(xmin, xmax, expander=0.1)
        ymin, ymax = mtrans.nonsingular(ymin, ymax, expander=0.1)

    padding = 1.0e-9 * (xmax - xmin)
    xmin -= padding
    xmax += padding

    n = len(x)
    polygon = np.zeros((6, 2), float)

    S = 1 / M_SQRT3
    mx = my = 0.99 * scale
    polygon[:, 0] = mx * np.array([-0.5 * S, 0.5 * S, 1.0 * S, 0.5 * S, -0.5 * S, -1.0 * S])
    polygon[:, 1] = my * np.array([0.5, 0.5, 0.0, -0.5, -0.5, 0.0])

    offsets = np.zeros((n, 2), float)
    offsets[:, 0] = x
    offsets[:, 1] = y

    collection = mcoll.PolyCollection(
        [polygon],
        edgecolors=edgecolors,
        linewidths=linewidths,
        offsets=offsets,
        transOffset=mtransforms.IdentityTransform(),
        offset_position="data"
    )

    if isinstance(norm, mcolors.LogNorm):
        if (z == 0).any():
            # make sure we have not zeros
            z += 1

    if norm is not None:
        if norm.vmin is None and norm.vmax is None:
            norm.autoscale(z)

    if norm is not None and not isinstance(norm, mcolors.Normalize):
        msg = "'norm' must be an instance of 'mcolors.Normalize'"
        raise ValueError(msg)

    collection.set_array(z)
    collection.set_cmap(cmap)
    collection.set_norm(norm)
    collection.set_alpha(alpha)
    collection.update(kwargs)

    if vmin is not None or vmax is not None:
        collection.set_clim(vmin, vmax)
    else:
        collection.autoscale_None()

    corners = ((xmin, ymin), (xmax, ymax))
    axis.update_datalim(corners)
    axis.autoscale_view(tight=True)

    # add the collection last
    axis.add_collection(collection, autolim=False)
    return collection


def transform_to_pixel(wcs, coords, ref=0):
    # We need only the WL axis
    coords_m = [coords, np.zeros_like(coords)]
    r = wcs.wcs_world2pix(np.transpose(coords_m), ref)
    first_col = r[:, 0].astype('int')
    return first_col


def is_inside(shape, *coords):

    for coord in coords:
        for shape1d, crds in zip(shape, coord):
            if not is_inside_1d(shape1d, crds):
                return False
    return True


def is_inside_1d(shape1d, coords):
    for coord in coords:
        if coord < 0:
            return False
        if coord >= shape1d:
            return False
    return True


def extract_region(wcs_wl, region, rssdata):
    cut1, cut2 = transform_to_pixel(wcs_wl, region)

    if is_inside_1d(rssdata.shape[1], [cut1, cut2]):
        z = rssdata[:, cut1:cut2].mean(axis=1)
        return z
    else:
        raise ValueError('region to average is outside image')


def extract_column(wcs_wl, col, rssdata):
    cut, = transform_to_pixel(wcs_wl, [col])

    if is_inside_1d(rssdata.shape[1], [cut]):
        # cut1, cut2 = args.average_region
        z = rssdata[:, cut]
        return z
    else:
        raise ValueError('column is outside image')


def extract_zval(rssdata, wcs_wl, coordinate_type, column=None, average_region=None, continuum_region=None):

    if continuum_region is None and average_region is None:
        raise ValueError('continuum_region and average_region are None')

    if rssdata.ndim == 1:
        # data is 1d
        print('RSS is 1D')
        zval = rssdata
    elif rssdata.ndim == 2:
        print('world coordinates in', coordinate_type)
        if column is not None:
            print('use column', column)
            zval = extract_column(wcs_wl, column, rssdata)
        else:
            print('compute average in', average_region)
            zval = extract_region(wcs_wl, average_region, rssdata)
        # Subtract continuum
        if continuum_region:
            print('compute continuum in', continuum_region)
            z0 = extract_region(wcs_wl, continuum_region, rssdata)
            zval -= z0
    else:
        raise ValueError('image must be 1D or 2D')
    return zval


def main(argv=None):
    """Process command line"""

    import argparse
    import json

    import matplotlib.pyplot as plt
    import astropy.io.fits as fits
    from astropy.wcs import WCS
    from astropy.visualization import simple_norm
    import matplotlib.transforms as mtransforms

    import megaradrp.datamodel as dm
    from megaradrp.processing.wcs import update_wcs_from_ipa, compute_pa_from_ipa

    try:
        from megaradrp.processing.cube import create_cube_from_rss
        has_contours = True
    except ImportError:
        create_cube_from_rss = None
        has_contours = False

    parser = argparse.ArgumentParser(description='Display MEGARA RSS images')
    parser.add_argument('--wcs-grid', action='store_true',
                        help='Display WCS grid')
    parser.add_argument('--wcs-pa-from-header', action='store_true',
                        help="Use PA angle from PC keys", dest='pa_from_header')
    parser.add_argument('--average-region', nargs=2, default=[1000, 3000],
                        type=int, help='Region of the RSS averaged on display')
    parser.add_argument('--extname', '-e', default='PRIMARY',
                        help='Name of the extension used')
    parser.add_argument('--column', '-c', type=int,
                        help='Column of the RSS on display')
    parser.add_argument('--continuum-region', nargs=2,
                        type=int,
                        help='Region of the RSS used for continuum subtraction')
    parser.add_argument('--coordinate-type', choices=['pixel', 'wcs'],
                        default='pixel',
                        help='Types of coordinates used')
    parser.add_argument('--colormap', type=plt.get_cmap,
                        help='Name of a valid matplotlib colormap')
    parser.add_argument('--plot-sky', action='store_true',
                        help='Plot SKY bundles')
    parser.add_argument('--plot-nominal-config', action='store_true',
                        help='Plot nominal configuration, do not use the header')
    parser.add_argument('--hide-values', action='store_true',
                        help='Do not show values out of range')
    parser.add_argument('--title', help='Title of the plot')
    parser.add_argument('--label', help='Legend of the colorbar')
    parser.add_argument('--hex-size', type=float,
                        help='Size of the hexagons (default is {})'.format(SCALE),
                        default=SCALE)
    parser.add_argument('--hex-rel-size', type=float,
                        help='Scale the size of hexagons by a factor',
                        default=1.0)
    parser.add_argument('--min-cut', type=float,
                        help='Inferior cut level')
    parser.add_argument('--max-cut', type=float,
                        help='Superior cut level')
    parser.add_argument('--percent', type=float,
                        help='Compute cuts using percentiles')
    parser.add_argument('--stretch',
                        choices=['linear', 'sqrt', 'power', 'log', 'asinh'],
                        default='linear',
                        help='Name of the strech method used for display'
                        )
    if has_contours:
        parser.set_defaults(has_contours=True)
        group_c = parser.add_argument_group('contouring')
        group_c.add_argument('--contour-pixel-size', type=float, default=0.4,
                             help="Pixel size in arc seconds for image reconstruction")
        group_c.add_argument('--contour-levels',
                             help="Contour levels")
        group_c.add_argument('--contour', action='store_true',
                             help="Draw contours")
        group_c.add_argument('--contour-image',
                             help="Image for computing contours")
        group_c.add_argument('--contour-image-column', type=int,
                             help='Column of image used for contouring')
        group_c.add_argument('--contour-image-save',
                             help='Save image used for contouring')
        group_c.add_argument('--contour-image-region', nargs=2, default=[1000, 3000],
                             type=int, help='Region of the image used for contouring')
        group_c.add_argument('--contour-is-density', action='store_true',
                             help='The data is a magnitude that does not require scaling')

    else:
        parser.set_defaults(has_contours=False)

    parser.add_argument('rss', metavar='RSS', nargs='+',
                        help='RSS images to process')

    args = parser.parse_args(args=argv)

    for fname in args.rss:
        with fits.open(fname) as img:

            datamodel = dm.MegaraDataModel()
            if args.plot_nominal_config:
                insmode = img['FIBERS'].header['INSMODE']
                fiberconf = datamodel.get_fiberconf_default(insmode)
            else:
                fiberconf = datamodel.get_fiberconf(img)
            plot_mask = np.ones((fiberconf.nfibers,), dtype=np.bool)
            if not args.plot_sky:
                skyfibers = fiberconf.sky_fibers()
                skyfibers.sort()
                skyfibers_idx = [(fibid - 1) for fibid in skyfibers]
                plot_mask[skyfibers_idx] = False

            x = np.empty((fiberconf.nfibers,))
            y = np.empty((fiberconf.nfibers,))
            # Key is fibid
            for _, fiber in sorted(fiberconf.fibers.items()):
                idx = fiber.fibid - 1
                x[idx] = fiber.x
                y[idx] = fiber.y

            extname = args.extname
            if args.coordinate_type == 'wcs':
                # read WCS from extname
                wcs_wl = WCS(img[extname].header)
            else:
                wcs_wl = WCS(naxis=len(img[extname].shape))

            rssdata = np.squeeze(img[extname].data)

            zval = extract_zval(rssdata,
                                wcs_wl,
                                args.coordinate_type,
                                args.column,
                                args.average_region,
                                args.continuum_region
                                )
            fig = plt.figure()
            
            x = x[plot_mask]
            y = y[plot_mask]
            zval = zval[plot_mask]
            projection = None

            if args.wcs_grid:
                print('display wcs grid')
                if not args.pa_from_header:
                    print('compute PA from IPA')
                    ipa = img[extname].header['IPA']
                    pa = compute_pa_from_ipa(ipa)
                    print('IPA is', ipa, ' PA is', pa)
                    hdr_fibers = update_wcs_from_ipa(img['FIBERS'].header, pa)
                else:
                    hdr_fibers = img['FIBERS'].header

                projection = WCS(hdr_fibers)

            # plt.subplots_adjust(hspace=0.5)
            ax = fig.add_axes([0.15, 0.1, 0.8, 0.8], projection=projection)

#            ax.set_xlim([-6.5, 6.5])
#            ax.set_ylim([-6, 6])

            if args.wcs_grid:
                ax.coords.grid(color='black', alpha=1.0, linestyle='solid')

            norm = simple_norm(zval, args.stretch,
                               min_cut=args.min_cut,
                               max_cut=args.max_cut,
                               percent=args.percent
                               )

            if args.hide_values:
                zdisp = zval.copy()
                if norm.vmin is not None:
                    zdisp[zval < norm.vmin] = np.nan
                if norm.vmax is not None:
                    zdisp[zval > norm.vmax] = np.nan
            else:
                zdisp = zval
            scale = args.hex_rel_size * args.hex_size
            col = hexplot(ax, x, y, zdisp, scale=scale, cmap=args.colormap, norm=norm)

            if args.title is not None:
                ax.set_title(args.title)

            cb = fig.colorbar(col)
            if args.label is not None:
                cb.set_label(args.label)

            if args.has_contours and args.contour:
                target_scale_arcsec = args.contour_pixel_size

                if args.contour_image is not None:
                    with fits.open(args.contour_image) as hdu_con:
                        extname = args.extname

                        if args.coordinate_type == 'wcs':
                            # read WCS from extname
                            wcs_wl_c = WCS(hdu_con[extname].header)
                        else:
                            wcs_wl_c = WCS(naxis=len(hdu_con[extname].shape))

                        rssdata_con = np.squeeze(hdu_con[extname].data)

                        zval_c = extract_zval(rssdata_con,
                                              wcs_wl_c,
                                              args.coordinate_type,
                                              column=args.contour_image_column,
                                              average_region=args.contour_image_region,
                                              continuum_region=None)
                else:
                    zval_c = zval

                # Build synthetic rss... for reconstruction
                primary = fits.PrimaryHDU(data=zval_c[:, np.newaxis], header=img[extname].header)
                synt = fits.HDUList([primary, img['FIBERS']])

                if args.contour_image_save is not None:
                    synt.writeto(args.contour_image_save)

                conserve_flux = not args.contour_is_density
                s_cube = create_cube_from_rss(synt, target_scale_arcsec, conserve_flux=conserve_flux)
                cube_wcs = WCS(s_cube[0].header).celestial
                px, py = cube_wcs.wcs.crpix
                interp = np.squeeze(s_cube[0].data)
                td = mtransforms.Affine2D().translate(-px, -py).scale(target_scale_arcsec, target_scale_arcsec)
                tt_d = td + ax.transData
                # im = ax.imshow(interp, alpha=0.9, cmap='jet', transform=tt_d)
                # im = ax.imshow(interp, alpha=0.9, cmap='jet', transform=ax.get_transform(cube_wcs))
                if args.contour_levels is not None:
                    levels = json.loads(args.contour_levels)
                    mm = ax.contour(interp, levels, transform=tt_d)
                else:
                    mm = ax.contour(interp, transform=tt_d)
                print('contour levels', mm.levels)

            plt.show()


if __name__ == '__main__':
    main()



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import datetime

from astropy.io import fits
import numpy
from numina.processing import Corrector
from numina.frame.utils import copy_img


_logger = logging.getLogger(__name__)


class CommonFlatCorrector(Corrector):
    """A Node that corrects from fiber flat."""

    def __init__(self, flatdata, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        super(CommonFlatCorrector, self).__init__(datamodel=datamodel,
                                                 calibid=calibid,
                                                 dtype=dtype)
        if isinstance(flatdata, fits.HDUList):
            self.corr = flatdata[0].data
        elif isinstance(flatdata, fits.ImageHDU):
            self.corr = flatdata.data
        else:
            self.corr = numpy.asarray(flatdata)

        self.corrmean = self.corr.mean()
        self.flattag = 'flat'

    def run(self, img):
        imgid = self.get_imgid(img)
        _logger.debug('correct %s in image %s', self.flattag, imgid)

        # Avoid nan values when divide
        my_mask = self.corr == 0.0
        self.corr[my_mask] = 1.0

        img['primary'].data /= self.corr
        hdr = img['primary'].header

        self.header_update(hdr, imgid)

        return img

    def header_update(self, hdr, imgid):
        hdr['NUM-FLT'] = self.calibid
        hdr['history'] = 'Flat correction {}'.format(imgid)
        hdr['history'] = 'Flat correction time {}'.format(datetime.datetime.utcnow().isoformat())


class FiberFlatCorrector(CommonFlatCorrector):
    """A Node that corrects from fiber flat."""

    def __init__(self, fiberflat, datamodel=None, dtype='float32'):
        calibid = datamodel.get_imgid(fiberflat)
        super(FiberFlatCorrector, self).__init__(
            flatdata=fiberflat,
            datamodel=datamodel,
            calibid=calibid,
            dtype=dtype
        )
        self.flattag = 'fiberflat'

    def header_update(self, hdr, imgid):
        hdr['NUM-FIBF'] = self.calibid
        hdr['history'] = 'Fiber flat correction {}'.format(imgid)
        hdr['history'] = 'Fiber flat correction time {}'.format(datetime.datetime.utcnow().isoformat())


class Splitter(Corrector):
    """Make a copy of its input"""
    def __init__(self):
        super(Splitter, self).__init__()
        self.out = None

    def run(self, img):
        self.out = self.copy_img(img)
        return img

    def copy_img(self, img):
        return copy_img(img)


class FlipLR(Corrector):
    """Make a copy of its input"""
    def __init__(self):
        super(FlipLR, self).__init__()

    def run(self, img):
        img[0].data = numpy.fliplr(img[0].data)
        return img



#
# Copyright 2011-2018 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging
import datetime

from astropy.io import fits
import numpy

from numina.processing import Corrector


_logger = logging.getLogger('numina.processing')


class TwilightCorrector(Corrector):

    """A Node that corrects from twilight flat."""

    def __init__(self, twilight, datamodel=None, calibid='calibid-unknown', dtype='float32'):

        if isinstance(twilight, fits.HDUList):
            self.corr = twilight[0].data
            calibid = datamodel.get_imgid(twilight)
        elif isinstance(twilight, fits.ImageHDU):
            self.corr = twilight.data
        else:
            self.corr = numpy.asarray(twilight)

        super(TwilightCorrector, self).__init__(datamodel=datamodel,
                                                calibid=calibid,
                                                dtype=dtype)

        self.corrmean = self.corr.mean()
        self.flattag = 'twilight'

    def run(self, img):
        imgid = self.get_imgid(img)
        cap = self.flattag.capitalize()
        _logger.debug('correct from %s in image %s', cap, imgid)

        # Avoid nan values when divide
        my_mask = self.corr == 0.0
        self.corr[my_mask] = 1.0

        img[0].data /= self.corr
        hdr = img['primary'].header

        self.header_update(hdr, imgid)

        return img

    def header_update(self, hdr, imgid):
        hdr['NUM-TWIF'] = self.calibid
        cap = self.flattag.capitalize()
        now = datetime.datetime.utcnow().isoformat()
        hdr['history'] = '{} flat correction {}'.format(cap, imgid)
        hdr['history'] = '{} flat correction time {}'.format(cap, now)



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of Megara DRP
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import numpy
import scipy.interpolate as ii
from astropy import units as u
from numina.instrument.simulation.efficiency import Efficiency


class MegaraVPH(object):

    def __init__(self, name, setup, vphtable, resolution, transmission=None, conf=None):
        self.SAMPLING = 9.0

        self.name = name
        self.setup = setup
        self._res = resolution

        rr = numpy.loadtxt(vphtable)
        r1 = rr[:,0] # Position in the pseudoslit
        r2 = rr[:,1] # WL
        r3 = rr[:,2] # X position
        r4 = rr[:,3] # Y position

        self.wlmin = rr[:,1].min()
        self.wlmax = rr[:,1].max()

        # Bivariate interpolations
        self.ps_wl_x = ii.SmoothBivariateSpline(r1, r2, r3)
        self.ps_wl_y = ii.SmoothBivariateSpline(r1, r2, r4)

        self.ps_x_wl = ii.SmoothBivariateSpline(r1, r3, r2)
        self.ps_y_wl = ii.SmoothBivariateSpline(r1, r4, r2)

        # extreme values for interpolation

        self.wlmin_in = 0.98 * self.wlmin
        self.wlmax_in = 1.02 * self.wlmax

        if transmission is None:
            self._transmission = Efficiency()
        else:
            self._transmission = transmission

        if conf is None:
            conf = {}

        self.wl_range = conf.get('wl_range', [0.0, 0.0, 0.0])

    def distortion(self):
        pass

    def resolution(self, wl):
        return  self._res.response(wl)

    def config_info(self):
        return {'name': self.name, 'setup': self.setup, 'wl_range': self.wl_range}

    def wltable_interp(self):
        res_in = (self.wlmax/ self.resolution(self.wlmax)) / self.SAMPLING
        return numpy.arange(self.wlmin_in, self.wlmax_in, res_in) * u.micron

    def transmission(self, wl):
        return self._transmission.response(wl)


class DummyVPH(object):

    def __init__(self, name):
        self.SAMPLING = 9.0
        self.minwl = 3653.0
        self.maxwl = 4386.0
        self.wlmin_in = 0.3596
        self.wlmax_in = 0.4437
        self.res = 6028.0
        self.name = name


    def resolution(self, wl):
        # This is as VPH405_LR_res
        return  self.res * numpy.ones_like(wl)

    def config_info(self):
        return {'name': self.name}

    def wltable_interp(self):
        res_in = (self.wlmax_in / self.res) / self.SAMPLING
        return numpy.arange(self.wlmin_in, self.wlmax_in, res_in)

    def transmission(self, wl):
        return numpy.ones_like(wl)





import math

import numpy
import pytest

from ..wcs import compute_pa_from_ipa
from ..wcs import update_wcs_from_ipa


@pytest.mark.parametrize("ipa, pa", [(0, 163.854), (196.146,0), (90, 253.854)])
def test_pa_from_ipa1(ipa, pa):

    res = compute_pa_from_ipa(ipa)
    assert numpy.isclose([res % 360], [pa])


@pytest.mark.parametrize(
    "ipa, iaa, pa", [
        (0, -163.854, 163.854), (196.146, -163.854, 0), (90, -163.854, 253.854)
    ]
)
def test_pa_from_ipa2(ipa, iaa, pa):

    res = compute_pa_from_ipa(ipa, iaa)
    assert numpy.isclose([res % 360], [pa])


@pytest.mark.parametrize(
    "cdelt1", [
        0.1, -0.1
    ]
)
@pytest.mark.parametrize(
    "pa", [0.0, 45.0, 120.0,
    ]
)
def test_update_wcs_from_ipa(cdelt1, pa):

    hdr = {}
    hdr['CDELT1'] = cdelt1
    hdr = update_wcs_from_ipa(hdr, pa)
    pa_rad = numpy.deg2rad(pa)
    cos_a = math.cos(pa_rad)
    sin_a = math.sin(pa_rad)
    res = [[cos_a, sin_a], [-sin_a, cos_a]]
    val = [[hdr['PC1_1'], hdr['PC1_2']], [hdr['PC2_1'], hdr['PC2_2']]]

    assert hdr['CDELT1'] == -abs(cdelt1)
    assert numpy.allclose(val, res)

